from ross import SealElement, HolePatternSeal, LabyrinthSeal
from ross.units import check_units
from prettytable import PrettyTable
from plotly import graph_objects as go
from plotly.subplots import make_subplots

__all__ = ["HybridSeal"]


class HybridSeal(SealElement):
    """Hybrid seal - Compressible flow model with rotordynamic coefficients.

    This class provides a model for hybrid seals that combine a hole-pattern seal
    (damping section) with a labyrinth seal (throttling section). The model iteratively
    determines the interface pressure between the two seal stages by matching their leakage
    rates, then combines their rotordynamic coefficients.

    **Theoretical Approach:**

    1. **Iterative Pressure Matching**:
       - Uses bisection method to find intermediate pressure between seal stages
       - Ensures mass conservation: leakage into hole-pattern = leakage from labyrinth
       - Convergence criterion based on relative leakage difference

    2. **Combined Rotordynamic Coefficients**:
       - Direct and cross-coupled stiffness (K), damping (C), and mass (M) coefficients
       - Series combination: forces from both seals are added
       - Frequency-dependent coefficients for each operating speed

    Parameters
    ----------

    **Common Parameters (both seal stages):**

    n : int
        Node in which the hybrid seal will be located.
    shaft_radius : float, pint.Quantity
        Radius of shaft (m).
    inlet_pressure : float
        Total inlet pressure at labyrinth entrance (Pa).
    outlet_pressure : float
        Final outlet pressure at hole-pattern exit (Pa).
    inlet_temperature : float
        Inlet temperature (K).
    frequency : float, list, pint.Quantity,
        Shaft rotational speed(s) (rad/s).
        Can be a single value or list of frequencies.
    gas_composition : dict, optional
        Gas composition as a dictionary {component: molar_fraction}.
        Example: {"Nitrogen": 0.79, "Oxygen": 0.21} for air.
    molar : float, pint.Quantity, optional
        Molecular mass (kg/kgmol). For Air: molar=28.97 kg/kgmol.
        Required if gas_composition is None. Default is None.
    gamma : float, optional
        Gas constant gamma (Cp/Cv). For Air: gamma=1.4.
        Required if gas_composition is None. Default is None.

    **Hole-Pattern Seal Parameters (downstream damping stage):**

    hole_pattern_parameters : dict
        Other parameters for hole-pattern seal with the following keys:
        - radial_clearance : float, pint.Quantity
            Seal clearance (m).
        - length : float, pint.Quantity
            Length of the seal (m).
        - roughness : float
            E / D (roughness / diameter) of the shaft.
        - cell_length : float, pint.Quantity
            Typical length of a cell in the axial direction (m).
        - cell_width : float, pint.Quantity
            Typical length of a cell in the azimuthal direction (m).
        - cell_depth : float, pint.Quantity
            Depth of a cell (m).
        - b_suther : float, optional
            b coefficient for the Sutherland viscosity model.
            Required if gas_composition is None. Default is None.
        - s_suther : float, optional
            s coefficient for the Sutherland viscosity model.
            Required if gas_composition is None. Default is None.
        - preswirl : float, optional
            Ratio of the circumferential velocity of the gas to the surface velocity of the shaft.
            Default is 0.0.
        - entr_coef : float, optional
            Entrance loss coefficient.
            Default is 0.1.
        - exit_coef : float, optional
            Exit loss coefficient.
            Default is 0.5.
        - whirl_ratio : float, optional
            Ratio of whirl frequency to rotational speed.
            Default is 1.0.
        - nz : int, optional
            Number of discretization points in the axial direction.
            Default is 80.
        - max_iterations : int, optional
            Maximum number of iterations for basic state calculation.
            Default is 180.
        - tolerance : float, optional
            Tolerance of the solution expressed as a percentage of the pressure differential
            across the seal. Default is 0.0001.
        - first_step_size : float, optional
            Initial step for the solution method. It should not be more than 0.01.
            Default is 0.01.
        - rlx_factor : float, optional
            Relaxation factor. Should be smaller than 0.1.
            Default is 0.1.

    **Labyrinth Seal Parameters (upstream throttling stage):**

    labyrinth_parameters : dict
        Other parameters for labyrinth seal with the following keys:
        - radial_clearance : float, pint.Quantity
            Nominal radial clearance (m).
        - n_teeth : int
            Number of teeth (throttlings). Needs to be <= 30.
        - pitch : float, pint.Quantity
            Seal pitch (length of land) or axial cavity length (m).
        - tooth_height : float, pint.Quantity
            Height of seal strip (m).
        - tooth_width : float, pint.Quantity
            Thickness of throttle (tip-width) (m), used in mass flow calculation.
        - seal_type : str
            Indicates where labyrinth teeth are located.
            Specify 'rotor' if teeth are on rotor only.
            Specify 'stator' if teeth are on stator only.
            Specify 'inter' for interlocking type labyrinths.
        - preswirl : float
            Inlet swirl velocity ratio. Positive values for swirl with shaft rotation
            and negative values for swirl against shaft rotations.
        - tz : list of float, optional
            Temperature at states: [T_state1, T_state2] (deg K).
            Required if gas_composition is None.
            Default is None.
        - muz : list of float, optional
            Dynamic viscosity at states: [mu_state1, mu_state2] (kg/(m·s)).
            Required if gas_composition is None.
            Default is None.
        - analz : str, optional
            Indicates what will be analysed.
            Specify "FULL" for rotordynamic calculation and leakage analysis.
            Specify "LEAKAGE" for leakage analysis only.
            Default is "FULL".
        - nprt : int, optional
            Number of parameters to be printed in the output: 1 maximum, 5 minimum.
            Default is 1.
        - iopt1 : int, optional
            Use or no use of tangential momentum parameters introduced by Jenny and Kanki.
            Specify value 0 to not use parameters.
            Specify value 1 to use parameters.
            Default is 0.

    **Hybrid Seal Control Parameters:**

    tolerance : float, optional
        Tolerance for pressure matching. Default is 1e-6.
    max_iterations : int, optional
        Maximum iterations for pressure matching. Default is 1e20.
    color : str, optional
        Color for element visualization. Default is "#787FF6".
    scale_factor : float, optional
        Scale factor for element drawing. Default is 0.75.
    kwargs : optional
        Additional keyword arguments passed to parent SealElement.

    Examples
    --------
    >>> from ross.seals.hybrid_seal import HybridSeal
    >>> from ross.units import Q_
    >>> gas_composition = {
    ...   "Nitrogen": 0.7812,
    ...   "Oxygen": 0.2096,
    ...   "Argon": 0.0092,
    ... }
    >>> holep_params = {
    ...   "radial_clearance": 0.0003,
    ...   "length": 0.04,
    ...   "roughness": 0.0001,
    ...   "cell_length": 0.003,
    ...   "cell_width": 0.003,
    ...   "cell_depth": 0.002,
    ...   "preswirl": 0.8,
    ...   "entr_coef": 0.5,
    ...   "exit_coef": 1.0,
    ... }
    >>> laby_params = {
    ...   "radial_clearance": Q_(0.25, "mm"),
    ...   "n_teeth": 10,
    ...   "pitch": Q_(3, "mm"),
    ...   "tooth_height": Q_(3, "mm"),
    ...   "tooth_width": Q_(0.15, "mm"),
    ...   "seal_type": "inter",
    ...   "preswirl": 0.9,
    ...   "tz": [300.0, 299.5],
    ...   "muz": [1.85e-05, 1.84e-05],
    ... }
    >>> hybrid = HybridSeal(
    ...   n=0,
    ...   shaft_radius=Q_(25, "mm"),
    ...   inlet_pressure=500000,
    ...   outlet_pressure=100000,
    ...   inlet_temperature=300.0,
    ...   frequency=Q_([2000, 3000, 5000], "RPM"),
    ...   gas_composition=gas_composition,
    ...   hole_pattern_parameters=holep_params,
    ...   labyrinth_parameters=laby_params,
    ... )
    >>> hybrid.seal_leakage  # doctest: +ELLIPSIS
    0.0348...
    """

    @check_units
    def __init__(
        self,
        n,
        shaft_radius,
        inlet_pressure,
        outlet_pressure,
        inlet_temperature,
        frequency,
        hole_pattern_parameters,
        labyrinth_parameters,
        gas_composition=None,
        molar=None,
        gamma=None,
        tolerance=1e-6,
        max_iterations=1e20,
        color="#787FF6",
        scale_factor=0.75,
        **kwargs,
    ):
        p_low = outlet_pressure
        p_high = inlet_pressure
        iteration = 0
        convergence_leakage = 1

        self.convergence_history = []
        self.pressure_history = []
        self.leakage_laby_history = []
        self.leakage_hole_history = []

        while convergence_leakage > tolerance and iteration < max_iterations:
            interface_pressure = (p_low + p_high) / 2

            holep = HolePatternSeal(
                n=n,
                inlet_pressure=inlet_pressure,
                outlet_pressure=interface_pressure,
                inlet_temperature=inlet_temperature,
                frequency=frequency,
                shaft_radius=shaft_radius,
                gas_composition=gas_composition,
                molar=molar,
                gamma=gamma,
                **hole_pattern_parameters,
            )

            laby = LabyrinthSeal(
                n=n,
                inlet_pressure=interface_pressure,
                outlet_pressure=outlet_pressure,
                inlet_temperature=inlet_temperature,
                frequency=frequency,
                shaft_radius=shaft_radius,
                gas_composition=gas_composition,
                molar=molar,
                gamma=gamma,
                **labyrinth_parameters,
            )

            hole_leakage = holep.seal_leakage[0]
            laby_leakage = laby.seal_leakage[0]

            convergence_leakage = abs(laby_leakage - hole_leakage) / hole_leakage

            if hole_leakage > laby_leakage:
                p_low = interface_pressure
            else:
                p_high = interface_pressure

            iteration += 1

            self.convergence_history.append(convergence_leakage)
            self.pressure_history.append(interface_pressure)
            self.leakage_laby_history.append(laby_leakage)
            self.leakage_hole_history.append(hole_leakage)

        self.laby = laby
        self.hole_pattern = holep
        self.interface_pressure = interface_pressure
        self.n_iterations = iteration

        coefficients_dict = {
            c: [l + h for l, h in zip(getattr(laby, c), getattr(holep, c))]
            for c in laby._get_coefficient_list()
        }

        super().__init__(
            n,
            frequency=frequency,
            seal_leakage=laby_leakage,
            color=color,
            scale_factor=scale_factor,
            **coefficients_dict,
            **kwargs,
        )

    def summary_results(self):
        """Print summary of hybrid seal analysis results.

        This method displays a comprehensive summary of the hybrid seal calculation,
        including convergence information, seal leakage, interface pressure.

        Returns
        -------
        table : PrettyTable object
            Formatted table of results.
        """
        # Convergence summary
        table = PrettyTable()
        table.field_names = ["Parameter", "Value"]
        table.align["Parameter"] = "l"
        table.align["Value"] = "r"

        table.add_row(["Number of Iterations", f"{self.n_iterations}"])
        table.add_row(["Final Convergence", f"{self.convergence_history[-1]:.6e}"])
        table.add_row(["Interface Pressure (Pa)", f"{self.interface_pressure:.2f}"])
        table.add_row(["Seal Leakage (kg/s)", f"{self.seal_leakage:.6e}"])

        return table

    def plot_convergence(self):
        """Plot convergence history.

        This method creates a unified figure with three subplots showing:
        convergence history, interface pressure evolution, and leakage comparison.

        Returns
        -------
        fig : Plotly graph_objects.Figure()
            Plotly figure object with combined subplots.
        """

        fig = make_subplots(
            rows=3,
            cols=1,
            subplot_titles=(
                "Convergence History",
                "Leakage Rate Convergence",
                "Interface Pressure Evolution",
            ),
            vertical_spacing=0.12,
            specs=[[{"type": "scatter"}], [{"type": "scatter"}], [{"type": "scatter"}]],
        )

        iterations = list(range(1, len(self.convergence_history) + 1))

        fig.add_trace(
            go.Scatter(
                x=iterations,
                y=self.convergence_history,
                mode="lines+markers",
                name="Convergence Error",
                line=dict(width=2),
                marker=dict(size=6),
                legendgroup="group1",
            ),
            row=1,
            col=1,
        )

        fig.add_trace(
            go.Scatter(
                x=iterations,
                y=self.leakage_hole_history,
                mode="lines+markers",
                name="Hole-Pattern Seal",
                line=dict(width=2),
                marker=dict(size=6),
                legendgroup="group2",
            ),
            row=2,
            col=1,
        )

        fig.add_trace(
            go.Scatter(
                x=iterations,
                y=self.leakage_laby_history,
                mode="lines+markers",
                name="Labyrinth Seal",
                line=dict(width=2),
                marker=dict(size=6),
                legendgroup="group2",
            ),
            row=2,
            col=1,
        )

        fig.add_trace(
            go.Scatter(
                x=iterations,
                y=self.pressure_history,
                mode="lines+markers",
                name="Interface Pressure",
                line=dict(width=2),
                marker=dict(size=6),
                legendgroup="group3",
            ),
            row=3,
            col=1,
        )

        fig.update_xaxes(title_text="Iteration", row=3, col=1)
        fig.update_xaxes(showticklabels=False, row=1, col=1)
        fig.update_xaxes(showticklabels=False, row=2, col=1)

        fig.update_yaxes(
            title_text="Relative Leakage Error",
            type="log",
            exponentformat="e",
            row=1,
            col=1,
        )
        fig.update_yaxes(
            title_text="Leakage Rate (kg/s)",
            exponentformat="e",
            row=2,
            col=1,
        )
        fig.update_yaxes(
            title_text="Pressure (Pa)",
            exponentformat="e",
            row=3,
            col=1,
        )

        fig.update_layout(
            height=900,
            showlegend=True,
            title_text="Hybrid Seal Analysis Results",
            title_x=0.5,
            title_font=dict(size=18),
        )

        return fig

    def plot_pressure_distribution(
        self, pressure_units="MPa", length_units="m", fig=None, **kwargs
    ):
        """Plot pressure distribution for the hybrid seal.

        Parameters
        ----------
        pressure_units : str, optional
            Units for the pressure.
            Default is "MPa".
        length_units : str, optional
            Units for the length.
            Default is "m".
        fig : go.Figure, optional
            Plotly figure object with the plot.
            Default is None.
        kwargs : optional
            Additional key word arguments can be passed to change the plot layout only
            (e.g. width=1000, height=800, ...).
            *See Plotly Python Figure Reference for more information.

        Returns
        -------
        fig : Plotly graph_objects.Figure()
            Plotly figure object.
        """

        if fig is None:
            fig = go.Figure()

        hole_data = self.hole_pattern.plot_pressure_distribution(
            pressure_units=pressure_units, length_units=length_units
        ).data[0]
        laby_data = self.laby.plot_pressure_distribution(
            pressure_units=pressure_units, length_units=length_units
        ).data[0]

        laby_data["x"] += hole_data["x"][-1]

        fig.add_trace(hole_data)
        fig.add_trace(laby_data)

        fig.update_xaxes(title_text=f"Axial Position ({length_units})")
        fig.update_yaxes(title_text=f"Pressure ({pressure_units})")

        fig.update_layout(
            title_text="Pressure Distribution - Hybrid Seal",
            title_font=dict(size=18),
            **kwargs,
        )

        return fig



"""Point mass module.

This module defines the PointMass class which will be used to link elements.
"""

import numpy as np
from plotly import graph_objects as go

from ross.element import Element
from ross.units import check_units

__all__ = ["PointMass"]


class PointMass(Element):
    """A point mass element.

    This class will create a point mass element.
    This element can be used to link other elements in the analysis.
    The mass provided to the element can be different on the x and y direction
    (e.g. different support inertia for x and y directions).

    Parameters
    ----------
    n: int
        Node which the bearing will be located in.
    m: float, pint.Quantity, optional
        Mass for the element.
    mx: float, pint.Quantity, optional
        Mass for the element on the x direction.
    my: float, pint.Quantity, optional
        Mass for the element on the y direction.
    mz: float, pint.Quantity, optional
        Mass for the element on the z direction.
    tag: str
        A tag to name the element
    scale_factor : float, optional
        The scale factor is used to scale the point mass drawing.
        Default is 1.
    color : str, optional
        A color to be used when the element is represented.
        Default is "DarkSalmon".

    Examples
    --------
    >>> p0 = PointMass(n=0, m=2)
    >>> p0.M()
    array([[2., 0., 0.],
           [0., 2., 0.],
           [0., 0., 2.]])
    >>> p1 = PointMass(n=0, mx=2, my=3, mz=4)
    >>> p1.M()
    array([[2., 0., 0.],
           [0., 3., 0.],
           [0., 0., 4.]])
    """

    @check_units
    def __init__(
        self,
        n=None,
        m=None,
        mx=None,
        my=None,
        mz=None,
        tag=None,
        scale_factor=1.0,
        color="DarkSalmon",
    ):
        self.n = n
        self.m = m

        if mx is None and my is None and mz is None:
            mx = float(m)
            my = float(m)
            mz = float(m)

        self.mx = float(mx)
        self.my = float(my)
        self.mz = 0.0 if mz is None else float(mz)
        self.tag = tag
        self.dof_global_index = None
        self.scale_factor = scale_factor
        self.color = color

    def __hash__(self):
        return hash(self.tag)

    def __eq__(self, other):
        """Equality method for comparisons.

        Parameters
        ----------
        other : obj
            parameter for comparison

        Returns
        -------
        True if other is equal to the reference parameter.
        False if not.

        Example
        -------
        >>> pointmass1 = point_mass_example()
        >>> pointmass2 = point_mass_example()
        >>> pointmass1 == pointmass2
        True
        """
        if self.__dict__ == other.__dict__:
            return True
        else:
            return False

    def __repr__(self):
        """Return a string representation of a point mass element.

        Returns
        -------
        A string representation of a point mass element object.

        Examples
        --------
        >>> point_mass = point_mass_example()
        >>> point_mass
        PointMass(n=0, mx=1.0, my=2.0, mz=3.0, tag='pointmass')
        """
        return (
            f"{self.__class__.__name__}"
            f"(n={self.n}, mx={self.mx:{0}.{5}},"
            f" my={self.my:{0}.{5}},"
            f" mz={self.mz:{0}.{5}},"
            f" tag={self.tag!r})"
        )

    def __str__(self):
        """Convert object into string.

        Returns
        -------
        The object's parameters translated to strings

        Example
        -------
        >>> print(PointMass(n=0, mx=2.5, my=3.25, mz=4.15, tag="PointMass"))
        Tag:              PointMass
        Node:             0
        Mass X dir. (kg): 2.5
        Mass Y dir. (kg): 3.25
        Mass Z dir. (kg): 4.15
        """
        return (
            f"Tag:              {self.tag}"
            f"\nNode:             {self.n}"
            f"\nMass X dir. (kg): {self.mx:{2}.{5}}"
            f"\nMass Y dir. (kg): {self.my:{2}.{5}}"
            f"\nMass Z dir. (kg): {self.mz:{2}.{5}}"
        )

    def M(self):
        """Mass matrix for an instance of a point mass element.

        This method will return the mass matrix for an instance of a point mass element.

        Returns
        -------
        M : np.ndarray
            A matrix of floats containing the values of the mass matrix.

        Examples
        --------
        >>> p1 = PointMass(n=0, mx=2, my=3, mz=4)
        >>> p1.M()
        array([[2., 0., 0.],
               [0., 3., 0.],
               [0., 0., 4.]])
        """
        mx = self.mx
        my = self.my
        mz = self.mz
        # fmt: off
        M = np.array([[mx,  0,  0],
                      [ 0, my,  0],
                      [ 0,  0, mz]])
        # fmt: on
        return M

    def C(self):
        """Damping matrix for an instance of a point mass element.

        This method will return the damping matrix for an instance of a point mass
        element.

        Returns
        -------
        C : np.ndarray
            A matrix of floats containing the values of the damping matrix.

        Examples
        --------
        >>> p1 = PointMass(n=0, mx=2, my=3, mz=4)
        >>> p1.C()
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])
        """
        C = np.zeros((3, 3))
        return C

    def K(self):
        """Stiffness matrix for an instance of a point mass element.

        This method will return the stiffness matrix for an instance of a point mass
        element.

        Returns
        -------
        K : np.ndarray
            A matrix of floats containing the values of the stiffness matrix.

        Examples
        --------
        >>> p1 = PointMass(n=0, mx=2, my=3, mz=4)
        >>> p1.K()
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])
        """
        K = np.zeros((3, 3))
        return K

    def G(self):
        """Gyroscopic matrix for an instance of a point mass element.

        This method will return the gyroscopic matrix for an instance of a point mass
        element.

        Returns
        -------
        G : np.ndarray
            A matrix of floats containing the values of the gyroscopic matrix.

        Examples
        --------
        >>> p1 = PointMass(n=0, mx=2, my=3, mz=4)
        >>> p1.G()
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])
        """
        G = np.zeros((3, 3))
        return G

    def dof_mapping(self):
        """Degrees of freedom mapping.

        Returns a dictionary with a mapping between degree of freedom and its index.

        Returns
        -------
        dof_mapping : dict
            A dictionary containing the degrees of freedom and their indexes.

        Examples
        --------
        The numbering of the degrees of freedom for each node.

        Being the following their ordering for a node:

        x_0 - horizontal translation
        y_0 - vertical translation
        z_0 - axial translation

        >>> p1 = PointMass(n=0, mx=2, my=3, mz=4)
        >>> p1.dof_mapping()
        {'x_0': 0, 'y_0': 1, 'z_0': 2}
        """
        return dict(x_0=0, y_0=1, z_0=2)

    def _patch(self, position, fig):
        """Point mass element patch.

        Patch that will be used to draw the point mass element using Plotly library.

        Parameters
        ----------
        position : float
            Position in which the patch will be drawn.
        fig : plotly.graph_objects.Figure
            The figure object which traces are added on.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            The figure object which traces are added on.
        """
        zpos, ypos, yc_pos = position

        radius = 0.005 * self.scale_factor

        customdata = [self.n, self.mx, self.my, self.mz]
        hovertemplate = (
            f"PointMass Node: {customdata[0]}<br>"
            + f"Mass (X): {customdata[1]:.3f}<br>"
            + f"Mass (Y): {customdata[2]:.3f}<br>"
            + f"Mass (Z): {customdata[2]:.3f}<br>"
        )

        fig.add_trace(
            go.Scatter(
                x=[zpos, zpos],
                y=np.add([ypos, -ypos], yc_pos),
                customdata=[customdata] * 2,
                text=hovertemplate,
                mode="markers",
                marker=dict(size=5.0, color=self.color),
                showlegend=False,
                name=self.tag,
                legendgroup="pointmass",
                hoverinfo="text",
                hovertemplate=hovertemplate,
                hoverlabel=dict(bgcolor=self.color),
            )
        )

        fig.add_shape(
            dict(
                type="circle",
                xref="x",
                yref="y",
                x0=zpos - radius,
                y0=ypos - radius + yc_pos,
                x1=zpos + radius,
                y1=ypos + radius + yc_pos,
                fillcolor=self.color,
                line_color="black",
            )
        )
        fig.add_shape(
            dict(
                type="circle",
                xref="x",
                yref="y",
                x0=zpos - radius,
                y0=-ypos - radius + yc_pos,
                x1=zpos + radius,
                y1=-ypos + radius + yc_pos,
                fillcolor=self.color,
                line_color="black",
            )
        )

        return fig


def point_mass_example():
    """Create an example of point mass element.

    This function returns an instance of a simple point mass.
    The purpose is to make available a simple model so that doctest
    can be written using it.

    Returns
    -------
    point_mass : ross.PointMass
        An instance of a point mass object.

    Examples
    --------
    >>> pointmass = point_mass_example()
    >>> pointmass.mx
    1.0
    """
    n = 0
    mx = 1.0
    my = 2.0
    mz = 3.0
    return PointMass(n=n, mx=mx, my=my, mz=mz, tag="pointmass")



import numpy as np
from re import search
from copy import deepcopy as copy
from collections.abc import Iterable
from scipy.integrate import cumulative_trapezoid as integrate

import ross as rs
from ross.gear_element import Mesh
from ross.rotor_assembly import Rotor
from ross.units import Q_, check_units

__all__ = ["MultiRotor"]


class MultiRotor(Rotor):
    """A class representing a multi-rotor system.

    This class creates a system comprising multiple rotors, with the specified
    driving rotor and driven rotor. For systems with more than two rotors,
    multiple multi-rotors can be nested.

    Parameters
    ----------
    driving_rotor : rs.Rotor
        The driving rotor object.
    driven_rotor : rs.Rotor
        The driven rotor object.
    coupled_nodes : tuple of int
        Tuple specifying the coupled nodes, where the first node corresponds to
        the driving rotor and the second node corresponds to the driven rotor.
    gear_mesh_stiffness : float, optional
        Directly specify the stiffness of the gear mesh.
        If not provided, it can be calculated automatically
        when using `GearElementTVMS` instead of `GearElement`.
    update_mesh_stiffness : bool, optional
        Applicable only when using `GearElementTVMS`.
        If True, the gear mesh stiffness is recalculated
        at each time step. If False, the maximum stiffness
        value is used throughout the simulation.
    square_varying_stiffness: boll, optional
        Set the square shape time varying mesh stiffness
    square_stiffness_amplitude_ratio: float, optional
        Ratio of stiffness amplitude based on the mean value of stiffness.
    orientation_angle : float, pint.Quantity, optional
        The angle between the line of gear centers and x-axis. Default is 0.0 rad.
    position : {'above', 'below'}, optional
        The relative position of the driven rotor with respect to the driving rotor
        when plotting the multi-rotor. Default is 'above'.
    tag : str, optional
        A tag to identify the multi-rotor. Default is None.

    Returns
    -------
    rotor : rs.Rotor
        The created multi-rotor object.

    Examples
    --------
    >>> import ross as rs
    >>> steel = rs.materials.steel
    >>> # Rotor 1:
    >>> L1 = [0.1, 4.24, 1.16, 0.3]
    >>> d1 = [0.3, 0.3, 0.22, 0.22]
    >>> shaft1 = [
    ...     rs.ShaftElement(
    ...         L=L1[i],
    ...         idl=0.0,
    ...         odl=d1[i],
    ...         material=steel,
    ...     )
    ...     for i in range(len(L1))
    ... ]
    >>> generator = rs.DiskElement(n=1, m=525.7, Id=16.1, Ip=32.2)
    >>> disk = rs.DiskElement(n=2, m=116.04, Id=3.115, Ip=6.23)
    >>> gear1 = rs.GearElement(
    ...     n=4, m=726.4, Id=56.95, Ip=113.9, n_teeth=328,
    ...     pitch_diameter=1.1, pr_angle=rs.Q_(22.5, 'deg'),
    ... )
    >>> bearing1 = rs.BearingElement(n=0, kxx=183.9e6, kyy=200.4e6, cxx=3e3)
    >>> bearing2 = rs.BearingElement(n=3, kxx=183.9e6, kyy=200.4e6, cxx=3e3)
    >>> rotor1 = rs.Rotor(shaft1, [generator, disk, gear1], [bearing1, bearing2],)

    >>> # Rotor 2:
    >>> L2 = [0.3, 5, 0.1]
    >>> d2 = [0.15, 0.15, 0.15]
    >>> shaft2 = [
    ...     rs.ShaftElement(
    ...         L=L2[i],
    ...         idl=0.0,
    ...         odl=d2[i],
    ...         material=steel,
    ...     )
    ...     for i in range(len(L2))
    ... ]
    >>> gear2 = rs.GearElement(
    ...     n=0, m=5, Id=0.002, Ip=0.004, n_teeth=23,
    ...     base_diameter=0.03567 * 2, pr_angle=rs.Q_(22.5, 'deg'),
    ... )
    >>> turbine = rs.DiskElement(n=2, m=7.45, Id=0.0745, Ip=0.149)
    >>> bearing3 = rs.BearingElement(n=1, kxx=10.1e6, kyy=41.6e6, cxx=3e3)
    >>> bearing4 = rs.BearingElement(n=3, kxx=10.1e6, kyy=41.6e6, cxx=3e3)
    >>> rotor2 = rs.Rotor(shaft2, [gear2, turbine], [bearing3, bearing4],)

    >>> # Multi rotor:
    >>> multi_rotor = rs.MultiRotor(
    ...     rotor1,
    ...     rotor2,
    ...     coupled_nodes=(4, 0),
    ...     gear_mesh_stiffness=1e8,
    ...     orientation_angle=0.0,
    ...     position="below"
    ... )
    >>> modal = multi_rotor.run_modal(speed=0)
    >>> modal.wd[0] # doctest: +ELLIPSIS
    74.160...
    """

    @check_units
    def __init__(
        self,
        driving_rotor,
        driven_rotor,
        coupled_nodes,
        gear_mesh_stiffness=None,
        update_mesh_stiffness=False,
        square_varying_stiffness=False,
        square_stiffness_amplitude_ratio=0,
        orientation_angle=0.0,
        position="above",
        tag=None,
    ):
        self.rotors = {"driving": driving_rotor, "driven": driven_rotor}
        self.orientation_angle = orientation_angle

        R1 = copy(driving_rotor)
        R2 = copy(driven_rotor)

        gear_1 = [
            elm
            for elm in R1.disk_elements
            if elm.n == coupled_nodes[0] and "GearElement" in type(elm).__name__
        ]
        gear_2 = [
            elm
            for elm in R2.disk_elements
            if elm.n == coupled_nodes[1] and "GearElement" in type(elm).__name__
        ]
        if len(gear_1) == 0 or len(gear_2) == 0:
            raise TypeError("Each rotor needs a GearElement in the coupled nodes!")
        else:
            gear_1 = gear_1[0]
            gear_2 = gear_2[0]

        self.update_mesh_stiffness = update_mesh_stiffness
        self.square_varying_stiffness = square_varying_stiffness
        self.square_stiffness_amplitude_ratio = square_stiffness_amplitude_ratio

        self.mesh = Mesh(
            gear_1,
            gear_2,
            gear_mesh_stiffness=gear_mesh_stiffness,
            square_varying_stiffness=self.square_varying_stiffness,
            square_stiffness_amplitude_ratio=square_stiffness_amplitude_ratio,
            orientation_angle=self.orientation_angle,
        )

        gear1_plot = next(
            (
                elm
                for elm in R1.plot_rotor().data
                if elm["legendgroup"] == "gears"
                and int(search(r"Gear Node: (\d+)", elm.text).group(1)) == gear_1.n
            ),
            None,
        )

        gear2_plot = next(
            (
                elm
                for elm in R2.plot_rotor().data
                if elm["legendgroup"] == "gears"
                and int(search(r"Gear Node: (\d+)", elm.text).group(1)) == gear_2.n
            ),
            None,
        )

        if position == "above":
            ymax = max(y for y in gear1_plot["y"] if y is not None)
            ymin = min(y for y in gear2_plot["y"] if y is not None)
            self.dy_pos = +abs(ymax - ymin)
        else:
            ymax = max(y for y in gear2_plot["y"] if y is not None)
            ymin = min(y for y in gear1_plot["y"] if y is not None)
            self.dy_pos = -abs(ymax - ymin)

        idx1 = R1.nodes.index(gear_1.n)
        idx2 = R2.nodes.index(gear_2.n)
        self.dz_pos = float(R1.nodes_pos[idx1] - R2.nodes_pos[idx2])

        R1_max_node = max([*R1.nodes, *R1.link_nodes])
        R2_min_node = min([*R2.nodes, *R2.link_nodes])
        d_node = 0
        if R1_max_node >= R2_min_node:
            d_node = R1_max_node + 1
            for elm in R2.elements:
                elm.n += d_node
                try:
                    elm.n_link += d_node
                except:
                    pass

        self.driven_nodes = [int(n + d_node) for n in R2.nodes]

        shaft_elements = [*R1.shaft_elements, *R2.shaft_elements]
        disk_elements = [*R1.disk_elements, *R2.disk_elements]
        bearing_elements = [*R1.bearing_elements, *R2.bearing_elements]
        point_mass_elements = [*R1.point_mass_elements, *R2.point_mass_elements]

        super().__init__(
            shaft_elements,
            disk_elements,
            bearing_elements,
            point_mass_elements,
            tag=tag,
        )

    def _set_tag(self, tag):
        """Set the tag for the current multi-rotor."""
        self.tag = tag or "MultiRotor 0"

    def _set_element_tag(self, elm, index):
        """Set a tag for the given element."""
        elm.tag = elm.get_class_name_prefix(index)

    def _fix_nodes_pos(self, index, node, nodes_pos_l):
        """Adjust node positions of the driven rotor."""
        if node < self.driven_nodes[0]:
            nodes_pos_l[index] = self.rotors["driving"].nodes_pos[
                self.rotors["driving"].nodes.index(node)
            ]
        elif node == self.driven_nodes[0]:
            nodes_pos_l[index] = self.rotors["driven"].nodes_pos[0] + self.dz_pos

    def _set_nodes(self, df_shaft):
        """Set nodes and nodes_pos lists."""
        self.nodes = [*self.rotors["driving"].nodes, *self.driven_nodes]

        R2_nodes_pos = [pos + self.dz_pos for pos in self.rotors["driven"].nodes_pos]
        self.nodes_pos = [*self.rotors["driving"].nodes_pos, *R2_nodes_pos]

        R2_center_line = [
            pos + self.dy_pos for pos in self.rotors["driven"].center_line_pos
        ]
        self.center_line_pos = [
            *self.rotors["driving"].center_line_pos,
            *R2_center_line,
        ]

    def _join_matrices(self, driving_matrix, driven_matrix):
        """Join matrices from the driving rotor and driven rotor to form the matrix of
        the coupled system.

        Parameters
        ----------
        driving_matrix : np.ndarray
            The matrix from the driving rotor.
        driven_matrix : np.ndarray
            The matrix from the driven rotor.

        Returns
        -------
        global_matrix : np.ndarray
            The combined matrix of the coupled system.
        """

        global_matrix = np.zeros((self.ndof, self.ndof))

        first_ndof = self.rotors["driving"].ndof
        global_matrix[:first_ndof, :first_ndof] = driving_matrix
        global_matrix[first_ndof:, first_ndof:] = driven_matrix

        return global_matrix

    def _unbalance_force(self, node, magnitude, phase, omega):
        """Calculate unbalance forces.

        This is an auxiliary function the calculate unbalance forces. It takes the
        force magnitude and phase and generate an array with complex values of forces
        on each degree degree of freedom of the given node.

        Parameters
        ----------
        node : int
            Node where the unbalance is applied.
        magnitude : float
            Unbalance magnitude (kg.m)
        phase : float
            Unbalance phase (rad)
        omega : list, float
            Array with the desired range of frequencies

        Returns
        -------
        F0 : list
            Unbalance force in each degree of freedom for each value in omega
        """
        speed = self.check_speed(node, omega)

        return super()._unbalance_force(node, magnitude, phase, speed)

    def unbalance_force_over_time(
        self, node, magnitude, phase, omega, t, return_all=False
    ):
        """Calculate unbalance forces for each time step.

        This auxiliary function calculates the unbalanced forces by taking
        into account the magnitude and phase of the force. It generates an
        array of force values at each degree of freedom for the specified
        nodes at each time step, while also considering a range of
        frequencies.

        Parameters
        ----------
        node : list, int
            Nodes where the unbalance is applied.
        magnitude : list, float
            Unbalance magnitude (kg.m) for each node.
        phase : list, float
            Unbalance phase (rad) for each node.
        omega : float, np.darray
            Constant velocity or desired range of velocities (rad/s).
        t : np.darray
            Time array (s).
        return_all : bool, optional
            If True, returns F0, theta, omega, and alpha.
            If False, returns only F0.
            Default is False.

        Returns
        -------
        F0 : np.ndarray
            Unbalance force at each degree of freedom for each time step.
        theta : np.ndarray
            Angular positions for each time step.
        omega : np.ndarray
            Angular velocities for each time step.
        alpha : np.ndarray
            Angular accelerations for each time step.
        """

        if not isinstance(omega, Iterable):
            omega = np.full_like(t, omega)

        theta = integrate(omega, t, initial=0)
        alpha = np.gradient(omega, t)

        F0 = np.zeros((self.ndof, len(t)))

        for i, n in enumerate(node):
            phi = phase[i] + theta

            w = self.check_speed(n, omega)

            Fx = magnitude[i] * ((w**2) * np.cos(phi) + alpha * np.sin(phi))
            Fy = magnitude[i] * ((w**2) * np.sin(phi) - alpha * np.cos(phi))

            F0[n * self.number_dof + 0, :] += Fx
            F0[n * self.number_dof + 1, :] += Fy

        if return_all:
            return F0, theta, omega, alpha
        else:
            return F0

    def check_speed(self, node, omega):
        """Adjusts the speed for the specified node based on the rotor configuration.

        This method checks if the given node belongs to the driven rotor.
        If so, the rotation speed is multiplied by the gear ratio.

        Parameters
        ----------
        node : int
            The node index where the speed check is being applied.
        omega : float or np.ndarray
            The original rotation speed of the driving rotor in rad/s.

        Returns
        -------
        speed : float or np.ndarray
            The adjusted rotation speed for the specified node.
        """

        speed = omega
        rotor = self.rotors["driving"]

        if node in self.driven_nodes:
            speed = -self.mesh.gear_ratio * omega
            rotor = self.rotors["driven"]

        if isinstance(rotor, MultiRotor):
            return rotor.check_speed(node, speed)

        return speed

    def coupling_matrix(self):
        """Coupling matrix of two coupled gears.

        coupling matrix according to:
        STRINGER, D. B. Geared Rotor Dynamic Methodologies for Advancing Prognostic Modeling
        Capabilities in Rotary-Wing Transmission Systems. Tese (Dissertation) — University of Virginia, Charlottesville, VA, 2008

        Returns
        -------
        coupling_matrix : np.ndarray
            Dimensionless coupling matrix of two coupled gears

        Examples
        --------
        >>> multi_rotor = two_shaft_rotor_example()
        >>> np.round(multi_rotor.coupling_matrix(),8)[:4, :4]
        array([[ 0.14644661,  0.35355339, -0.        , -0.        ],
               [ 0.35355339,  0.85355339, -0.        , -0.        ],
               [-0.        , -0.        ,  0.        ,  0.        ],
               [-0.        , -0.        ,  0.        ,  0.        ]])
        """

        pressure_angle = self.mesh.pressure_angle  # normal angle

        helix_angle = self.mesh.helix_angle
        helix_angle -= np.pi  # adjusted according to formulation

        cx = np.cos(pressure_angle) * np.cos(helix_angle)
        cy = np.sin(pressure_angle)
        cz = np.sin(pressure_angle) * np.sin(helix_angle)

        cp = np.cos(self.orientation_angle)
        sp = np.sin(self.orientation_angle)

        r1 = self.mesh.driving_gear.pitch_diameter / 2
        r2 = self.mesh.driven_gear.pitch_diameter / 2

        Kii = np.zeros((self.number_dof, self.number_dof))
        Kji = np.zeros((self.number_dof, self.number_dof))
        Kjj = np.zeros((self.number_dof, self.number_dof))

        # Kii
        Kii[0, 0] = (sp * cx + cp * cy) ** 2
        Kii[1, 0] = (sp * cx + cp * cy) * (sp * cy - cp * cx)
        Kii[2, 0] = cz * (sp * cx + cp * cy)
        Kii[3, 0] = sp * cz * r1 * (sp * cx + cp * cy)
        Kii[4, 0] = -cp * cz * r1 * (sp * cx + cp * cy)
        Kii[5, 0] = -cx * r1 * (cp**2 + sp**2) * (sp * cx + cp * cy)

        Kii[0, 1] = Kii[1, 0]
        Kii[1, 1] = (sp * cy - cp * cx) ** 2
        Kii[2, 1] = cz * (sp * cy - cp * cx)
        Kii[3, 1] = sp * cz * r1 * (sp * cy - cp * cx)
        Kii[4, 1] = -cp * cz * r1 * (sp * cy - cp * cx)
        Kii[5, 1] = -cx * r1 * (cp**2 + sp**2) * (sp * cy - cp * cx)

        Kii[0, 2] = Kii[2, 0]
        Kii[1, 2] = Kii[2, 1]
        Kii[2, 2] = cz**2
        Kii[3, 2] = sp * (cz**2) * r1
        Kii[4, 2] = -cp * (cz**2) * r1
        Kii[5, 2] = -cx * cz * r1 * (cp**2 + sp**2)

        Kii[0, 3] = Kii[3, 0]
        Kii[1, 3] = Kii[3, 1]
        Kii[2, 3] = Kii[3, 2]
        Kii[3, 3] = (sp * cz * r1) ** 2
        Kii[4, 3] = -cp * sp * (cz * r1) ** 2
        Kii[5, 3] = -sp * cx * cz * (r1**2) * (cp**2 + sp**2)

        Kii[0, 4] = Kii[4, 0]
        Kii[1, 4] = Kii[4, 1]
        Kii[2, 4] = Kii[4, 2]
        Kii[3, 4] = Kii[4, 3]
        Kii[4, 4] = (cp * cz * r1) ** 2
        Kii[5, 4] = cp * cx * cz * (r1**2) * (cp**2 + sp**2)

        Kii[0, 5] = Kii[5, 0]
        Kii[1, 5] = Kii[5, 1]
        Kii[2, 5] = Kii[5, 2]
        Kii[3, 5] = Kii[5, 3]
        Kii[4, 5] = Kii[5, 4]
        Kii[5, 5] = ((cx * r1) ** 2) * (cp**4 + 2 * (cp * sp) ** 2 + sp**4)

        # Kji
        Kji[0, 0] = -((sp * cx + cp * cy) ** 2)
        Kji[1, 0] = -(sp * cx + cp * cy) * (sp * cy - cp * cx)
        Kji[2, 0] = -cz * (sp * cx + cp * cy)
        Kji[3, 0] = -sp * cz * r2 * (sp * cx + cp * cy)
        Kji[4, 0] = cp * cz * r2 * (sp * cx + cp * cy)
        Kji[5, 0] = cx * r2 * (cp**2 + sp**2) * (sp * cx + cp * cy)

        Kji[0, 1] = Kji[1, 0]
        Kji[1, 1] = -((sp * cy - cp * cx) ** 2)
        Kji[2, 1] = -cz * (sp * cy - cp * cx)
        Kji[3, 1] = -sp * cz * r2 * (sp * cy - cp * cx)
        Kji[4, 1] = cp * cz * r2 * (sp * cy - cp * cx)
        Kji[5, 1] = cx * r2 * (cp**2 + sp**2) * (sp * cy - cp * cx)

        Kji[0, 2] = Kji[2, 0]
        Kji[1, 2] = Kji[2, 1]
        Kji[2, 2] = -(cz**2)
        Kji[3, 2] = -sp * (cz**2) * r2
        Kji[4, 2] = cp * (cz**2) * r2
        Kji[5, 2] = cx * cz * r2 * (cp**2 + sp**2)

        Kji[0, 3] = -sp * cz * r1 * (sp * cx + cp * cy)
        Kji[1, 3] = -sp * cz * r1 * (sp * cy - cp * cx)
        Kji[2, 3] = -sp * (cz**2) * r1
        Kji[3, 3] = -((sp * cz) ** 2) * r1 * r2
        Kji[4, 3] = cp * sp * (cz**2) * r1 * r2
        Kji[5, 3] = sp * cx * cz * r1 * r2 * (cp**2 + sp**2)

        Kji[0, 4] = cp * cz * r1 * (sp * cx + cp * cy)
        Kji[1, 4] = cp * cz * r1 * (sp * cy - cp * cx)
        Kji[2, 4] = cp * (cz**2) * r1
        Kji[3, 4] = cp * sp * (cz**2) * r1 * r2
        Kji[4, 4] = -((cp * cz) ** 2) * r1 * r2
        Kji[5, 4] = -cp * cx * cz * r1 * r2 * (cp**2 + sp**2)

        Kji[0, 5] = cx * r1 * (cp**2 + sp**2) * (sp * cx + cp * cy)
        Kji[1, 5] = cx * r1 * (cp**2 + sp**2) * (sp * cy - cp * cx)
        Kji[2, 5] = cx * cz * r1 * (cp**2 + sp**2)
        Kji[3, 5] = sp * cx * cz * r1 * r2 * (cp**2 + sp**2)
        Kji[4, 5] = -cp * cx * cz * r1 * r2 * (cp**2 + sp**2)
        Kji[5, 5] = -(cx**2) * r1 * r2 * (cp**4 + 2 * (cp * sp) ** 2 + sp**4)

        # Kjj
        Kjj[0, 0] = (sp * cx + cp * cy) ** 2
        Kjj[1, 0] = (sp * cx + cp * cy) * (sp * cy - cp * cx)
        Kjj[2, 0] = cz * (sp * cx + cp * cy)
        Kjj[3, 0] = sp * cz * r2 * (sp * cx + cp * cy)
        Kjj[4, 0] = -cp * cz * r2 * (sp * cx + cp * cy)
        Kjj[5, 0] = -cx * r2 * (cp**2 + sp**2) * (sp * cx + cp * cy)

        Kjj[0, 1] = Kjj[1, 0]
        Kjj[1, 1] = (sp * cy - cp * cx) ** 2
        Kjj[2, 1] = cz * (sp * cy - cp * cx)
        Kjj[3, 1] = sp * cz * r2 * (sp * cy - cp * cx)
        Kjj[4, 1] = -cp * cz * r2 * (sp * cy - cp * cx)
        Kjj[5, 1] = -cx * r2 * (cp**2 + sp**2) * (sp * cy - cp * cx)

        Kjj[0, 2] = Kjj[2, 0]
        Kjj[1, 2] = Kjj[2, 1]
        Kjj[2, 2] = cz**2
        Kjj[3, 2] = sp * (cz**2) * r2
        Kjj[4, 2] = -cp * (cz**2) * r2
        Kjj[5, 2] = -cx * cz * r2 * (cp**2 + sp**2)

        Kjj[0, 3] = Kjj[3, 0]
        Kjj[1, 3] = Kjj[3, 1]
        Kjj[2, 3] = Kjj[3, 2]
        Kjj[3, 3] = (sp * cz * r2) ** 2
        Kjj[4, 3] = -cp * sp * (cz * r2) ** 2
        Kjj[5, 3] = -sp * cx * cz * (r2**2) * (cp**2 + sp**2)

        Kjj[0, 4] = Kjj[4, 0]
        Kjj[1, 4] = Kjj[4, 1]
        Kjj[2, 4] = Kjj[4, 2]
        Kjj[3, 4] = Kjj[4, 3]
        Kjj[4, 4] = (cp * cz * r2) ** 2
        Kjj[5, 4] = cp * cx * cz * (r2**2) * (cp**2 + sp**2)

        Kjj[0, 5] = Kjj[5, 0]
        Kjj[1, 5] = Kjj[5, 1]
        Kjj[2, 5] = Kjj[5, 2]
        Kjj[3, 5] = Kjj[5, 3]
        Kjj[4, 5] = Kjj[5, 4]
        Kjj[5, 5] = ((cx * r2) ** 2) * (cp**4 + 2 * (cp * sp) ** 2 + sp**4)

        coupling_matrix = np.block([[Kii, Kji.T], [Kji, Kjj]])

        return coupling_matrix

    def M(self, frequency=None, synchronous=False):
        """Mass matrix for a multi-rotor.

        Parameters
        ----------
        synchronous : bool, optional
            If True a synchronous analysis is carried out.
            Default is False.

        Returns
        -------
        M0 : np.ndarray
            Mass matrix for the multi-rotor.

        Examples
        --------
        >>> multi_rotor = two_shaft_rotor_example()
        >>> multi_rotor.M(0)[:4, :4]
        array([[18.55298224,  0.        ,  0.        ,  0.        ],
               [ 0.        , 18.55298224,  0.        , -0.16179571],
               [ 0.        ,  0.        , 18.37831702,  0.        ],
               [ 0.        , -0.16179571,  0.        ,  0.10074262]])
        """

        if frequency is None:
            return self._join_matrices(
                self.rotors["driving"].M(synchronous=synchronous),
                self.rotors["driven"].M(synchronous=synchronous),
            )
        else:
            return self._join_matrices(
                self.rotors["driving"].M(frequency, synchronous),
                self.rotors["driven"].M(frequency * self.mesh.gear_ratio, synchronous),
            )

    def K(self, frequency):
        """Stiffness matrix for a multi-rotor.

        Parameters
        ----------
        frequency : float, optional
            Excitation frequency.

        Returns
        -------
        K0 : np.ndarray
            Stiffness matrix for the multi-rotor.

        Examples
        --------
        >>> multi_rotor = two_shaft_rotor_example()
        >>> multi_rotor.K(0)[:4, :4] / 1e10
        array([[ 4.7609372 ,  0.        ,  0.        ,  0.        ],
               [ 0.        ,  4.7625872 ,  0.        , -0.23712736],
               [ 0.        ,  0.        , 14.63196778,  0.        ],
               [ 0.        , -0.23712736,  0.        ,  0.09416119]])
        """

        K0 = self._join_matrices(
            self.rotors["driving"].K(frequency),
            self.rotors["driven"].K(frequency * self.mesh.gear_ratio),
        )

        dofs_1 = self.mesh.driving_gear.dof_global_index.values()
        dofs_2 = self.mesh.driven_gear.dof_global_index.values()
        dofs = [*dofs_1, *dofs_2]

        K0[np.ix_(dofs, dofs)] += self.coupling_matrix() * self.mesh.stiffness

        return K0

    def Ksdt(self):
        """Dynamic stiffness matrix for a multi-rotor.

        Stiffness matrix associated with the transient motion of the
        shaft and disks. For time-dependent analyses, this matrix needs to be
        multiplied by the angular acceleration. Therefore, the stiffness matrix
        of the driven rotor is scaled by the gear ratio before being combined
        with the driving rotor matrix.

        Returns
        -------
        Ksdt0 : np.ndarray
            Dynamic stiffness matrix for the multi-rotor.

        Examples
        --------
        >>> multi_rotor = two_shaft_rotor_example()
        >>> multi_rotor.Ksdt()[:6, :4]
        array([[  0.        , -74.43218395,   0.        ,   0.6202682 ],
               [  0.        ,   0.        ,   0.        ,   0.        ],
               [  0.        ,   0.        ,   0.        ,   0.        ],
               [  0.        ,   0.        ,   0.        ,   0.        ],
               [  0.        ,  -0.6202682 ,   0.        ,   0.08270243],
               [  0.        ,   0.        ,   0.        ,   0.        ]])
        """

        return self._join_matrices(
            self.rotors["driving"].Ksdt(),
            -self.mesh.gear_ratio * self.rotors["driven"].Ksdt(),
        )

    def C(self, frequency):
        """Damping matrix for a multi-rotor rotor.

        Parameters
        ----------
        frequency : float
            Excitation frequency.

        Returns
        -------
        C0 : np.ndarray
            Damping matrix for the multi-rotor.

        Examples
        --------
        >>> multi_rotor = two_shaft_rotor_example()
        >>> multi_rotor.C(0)[:4, :4] / 1e3
        array([[3., 0., 0., 0.],
               [0., 3., 0., 0.],
               [0., 0., 0., 0.],
               [0., 0., 0., 0.]])
        """

        return self._join_matrices(
            self.rotors["driving"].C(frequency),
            self.rotors["driven"].C(frequency * self.mesh.gear_ratio),
        )

    def G(self):
        """Gyroscopic matrix for a multi-rotor.

        For time-dependent analyses, this matrix needs to be multiplied by the
        rotor speed. Therefore, the gyroscopic matrix of the driven rotor is
        scaled by the gear ratio before being combined with the driving rotor matrix.

        Returns
        -------
        G0 : np.ndarray
            Gyroscopic matrix for the multi-rotor.

        Examples
        --------
        >>> multi_rotor = two_shaft_rotor_example()
        >>> multi_rotor.G()[:4, :4]
        array([[ 0.        ,  0.17162125,  0.        ,  0.1403395 ],
               [-0.17162125,  0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ,  0.        ],
               [-0.1403395 ,  0.        ,  0.        ,  0.        ]])
        """

        return self._join_matrices(
            self.rotors["driving"].G(),
            -self.mesh.gear_ratio * self.rotors["driven"].G(),
        )

    def _rotor_system_for_integrate(
        self, rotor, speed, t, reduce_matrix, forces, **kwargs
    ):
        """Build rotor system for integrate method."""
        if self.update_mesh_stiffness:
            # Check if speed is array
            speed_is_array = isinstance(speed, Iterable)
            if not speed_is_array:
                speed = np.full_like(t, speed)

            accel = np.gradient(speed, t)
            theta = integrate(speed, t, initial=0)

            # Assemble matrices
            M = reduce_matrix(kwargs.get("M", self.M()))
            C2 = reduce_matrix(kwargs.get("G", self.G()))
            K2 = reduce_matrix(kwargs.get("Ksdt", self.Ksdt()))

            def rotor_system(step, **current_state):
                C1 = reduce_matrix(self.C(speed[step]))

                # Update mesh stiffness
                self.mesh.stiffness = self.mesh.interpolate_stiffness(theta[step])
                K1 = reduce_matrix(self.K(speed[step]))

                return (
                    M,
                    C1 + C2 * speed[step],
                    K1 + K2 * accel[step],
                    forces(step, **current_state),
                )

            return rotor_system

        else:
            return super()._rotor_system_for_integrate(
                rotor, speed, t, reduce_matrix, forces, **kwargs
            )


def two_shaft_rotor_example():
    """Create a multi-rotor as example.

    This function returns an instance of two-shaft rotor system from Rao et al.
    This typical example is a turbo-alternator rotor system, which consists of
    a generator rotor, a turbine rotor and a spur gear pair connecting two rotors.
    Each rotor is supported by a pair of bearing two shaft elements, one disk and
    two simple bearings.

    The purpose of this is to make available a simple model so that doctest can
    be written using this.

    Returns
    -------
    An instance of a rotor object.

    References
    ----------
    Rao, J. S., Shiau, T. N., chang, J. R. (1998). Theoretical analysis of lateral
    response due to torsional excitation of geared rotors. Mechanism and Machine Theory,
    33 (6), 761-783. doi: 10.1016/S0094-114X(97)00056-6

    Examples
    --------
    >>> multi_rotor = two_shaft_rotor_example()
    >>> modal = multi_rotor.run_modal(speed=0)
    >>> np.round(modal.wd[:4])
    array([ 74.,  77., 112., 113.])
    """
    # A spur geared two-shaft rotor system.
    material = rs.Material(name="mat_steel", rho=7800, E=207e9, G_s=79.5e9)

    # Rotor 1
    L1 = [0.1, 4.24, 1.16, 0.3]
    d1 = [0.3, 0.3, 0.22, 0.22]
    shaft1 = [
        rs.ShaftElement(
            L=L1[i],
            idl=0.0,
            odl=d1[i],
            material=material,
            shear_effects=True,
            rotary_inertia=True,
            gyroscopic=True,
        )
        for i in range(len(L1))
    ]

    generator = rs.DiskElement(
        n=1,
        m=525.7,
        Id=16.1,
        Ip=32.2,
    )
    disk = rs.DiskElement(
        n=2,
        m=116.04,
        Id=3.115,
        Ip=6.23,
    )

    pressure_angle = Q_(22.5, "deg")

    gear1 = rs.GearElement(
        n=4,
        m=726.4,
        Id=56.95,
        Ip=113.9,
        n_teeth=328,
        base_diameter=0.5086 * 2,
        pr_angle=pressure_angle,
    )

    bearing1 = rs.BearingElement(n=0, kxx=183.9e6, kyy=200.4e6, cxx=3e3)
    bearing2 = rs.BearingElement(n=3, kxx=183.9e6, kyy=200.4e6, cxx=3e3)

    rotor1 = rs.Rotor(
        shaft1,
        [generator, disk, gear1],
        [bearing1, bearing2],
    )

    # Rotor 2
    L2 = [0.3, 5, 0.1]
    d2 = [0.15, 0.15, 0.15]
    shaft2 = [
        rs.ShaftElement(
            L=L2[i],
            idl=0.0,
            odl=d2[i],
            material=material,
            shear_effects=True,
            rotary_inertia=True,
            gyroscopic=True,
        )
        for i in range(len(L2))
    ]

    gear2 = rs.GearElement(
        n=0,
        m=5,
        Id=0.002,
        Ip=0.004,
        n_teeth=23,
        base_diameter=0.03567 * 2,
        pr_angle=pressure_angle,
    )

    turbine = rs.DiskElement(n=2, m=7.45, Id=0.0745, Ip=0.149)

    bearing3 = rs.BearingElement(n=1, kxx=10.1e6, kyy=41.6e6, cxx=3e3)
    bearing4 = rs.BearingElement(n=3, kxx=10.1e6, kyy=41.6e6, cxx=3e3)

    rotor2 = rs.Rotor(
        shaft2,
        [gear2, turbine],
        [bearing3, bearing4],
    )

    return rs.MultiRotor(
        rotor1,
        rotor2,
        coupled_nodes=(4, 0),
        gear_mesh_stiffness=1e8,
        orientation_angle=0.0,
        position="below",
    )



import os
import numpy as np
import pytest

from numpy.testing import assert_allclose

from ross.seals.holepattern_seal import HolePatternSeal
from ross.units import Q_

import ccp

# Detect REFPROP availability using ccp
# ccp is always available as a dependency, but REFPROP is optional
# Check if REFPROP version is available (not "n/a")
REFPROP_AVAILABLE = "REFPROP : n/a" not in ccp.__version__full


# Common test parameters
COMMON_PARAMS = {
    "n": 0,
    "frequency": Q_([5000], "RPM"),
    "length": 0.0254,
    "shaft_radius": 0.0751,
    "radial_clearance": 0.0004,
    "roughness": 0.00198,
    "cell_length": 0.001,
    "cell_width": 0.001,
    "cell_depth": 0.00229,
    "inlet_pressure": 1830000.0,
    "outlet_pressure": 823500.0,
    "inlet_temperature": 300.0,
    "preswirl": 1.0,
    "entr_coef": 0.1,
    "exit_coef": 0.5,
}

MANUAL_PARAMS = {
    "b_suther": 1.458e-6,
    "s_suther": 110.4,
    "molar": 29.0,
    "gamma": 1.4,
}

GAS_COMPOSITION = {
    "Nitrogen": 0.7812,
    "Oxygen": 0.2096,
    "Argon": 0.0092,
}


@pytest.fixture
def holepattern_manual():
    """HolePatternSeal with manual thermodynamic parameters."""
    seal = HolePatternSeal(**COMMON_PARAMS, **MANUAL_PARAMS)
    return seal


@pytest.fixture
def holepattern_gas_composition():
    """HolePatternSeal with gas_composition."""
    seal = HolePatternSeal(**COMMON_PARAMS, gas_composition=GAS_COMPOSITION)
    return seal


def test_holepattern_manual_coefficients(holepattern_manual):
    """Test HolePatternSeal coefficients with manual parameters.

    This test should always work regardless of REFPROP availability.
    """
    assert_allclose(holepattern_manual.kxx, 586228.88958017, rtol=1e-4)
    assert_allclose(holepattern_manual.kxy, 159741.17580073, rtol=1e-4)
    assert_allclose(holepattern_manual.kyx, -159741.17580073, rtol=1e-4)
    assert_allclose(holepattern_manual.kyy, 586228.88958017, rtol=1e-4)
    assert_allclose(holepattern_manual.cxx, 294.42942927, rtol=1e-4)
    assert_allclose(holepattern_manual.cxy, -27.16217997, rtol=1e-4)
    assert_allclose(holepattern_manual.cyx, 27.16217997, rtol=1e-4)
    assert_allclose(holepattern_manual.cyy, 294.42942927, rtol=1e-4)
    assert_allclose(holepattern_manual.seal_leakage, 0.6313559209954082, rtol=1e-4)


def test_holepattern_manual_attributes(holepattern_manual):
    """Test that manual parameters are correctly stored."""
    assert holepattern_manual.molar == 29.0
    assert holepattern_manual.gamma == 1.4
    assert holepattern_manual.b_suther == 1.458e-6
    assert holepattern_manual.s_suther == 110.4


def test_holepattern_gas_composition_creation(holepattern_gas_composition):
    """Test that HolePatternSeal can be created with gas_composition.

    This should work with both REFPROP and HEOS backends.
    """
    assert holepattern_gas_composition is not None
    assert hasattr(holepattern_gas_composition, "kxx")
    assert hasattr(holepattern_gas_composition, "seal_leakage")


def test_holepattern_gas_composition_derived_properties(holepattern_gas_composition):
    """Test that thermodynamic properties are auto-derived correctly."""
    # Check that properties were derived
    assert holepattern_gas_composition.molar is not None
    assert holepattern_gas_composition.gamma is not None
    assert holepattern_gas_composition.b_suther is not None
    assert holepattern_gas_composition.s_suther is not None

    # Molar mass should be close to air (28.97 g/mol)
    # Expected: ~28.96 kg/kgmol, tolerance: 2%
    assert_allclose(holepattern_gas_composition.molar, 28.97, rtol=0.02)

    # Gamma should be close to air (1.4)
    # Expected: ~1.4, tolerance: 3% (HEOS and REFPROP differ slightly)
    assert_allclose(holepattern_gas_composition.gamma, 1.4, rtol=0.03)

    # Sutherland coefficients should be in reasonable range for air
    # Expected: ~1.46e-6, tolerance: 25% (varies between HEOS and REFPROP)
    assert_allclose(holepattern_gas_composition.b_suther, 1.46e-6, rtol=0.25)
    # Expected: ~100, tolerance: 25% (varies between HEOS and REFPROP)
    assert_allclose(holepattern_gas_composition.s_suther, 100, rtol=0.25)


def test_holepattern_gas_composition_coefficients_range(holepattern_gas_composition):
    """Test that coefficients are in reasonable range.

    Results may differ slightly between REFPROP and HEOS, but should be
    within 1% of each other.
    """
    # Expected values with some tolerance for backend differences
    # REFPROP: kxx ≈ 606105, HEOS: kxx ≈ 605752
    # Use average of REFPROP and HEOS as expected value
    assert_allclose(holepattern_gas_composition.kxx[0], 606000, rtol=0.025)  # ±2.5%
    assert_allclose(abs(holepattern_gas_composition.kxy[0]), 160000, rtol=0.1)  # ±10%
    assert_allclose(holepattern_gas_composition.cxx[0], 300, rtol=0.05)  # ±5%
    assert_allclose(
        holepattern_gas_composition.seal_leakage[0], 0.634, rtol=0.02
    )  # ±2%


def test_holepattern_gas_composition_vs_manual_similarity():
    """Test that gas_composition and manual parameters give similar results.

    They won't be identical because:
    1. gas_composition auto-derives properties from real gas data
    2. Manual parameters use simplified air properties

    But they should be within ~5% of each other.
    """
    holepattern_gas = HolePatternSeal(**COMMON_PARAMS, gas_composition=GAS_COMPOSITION)
    holepattern_man = HolePatternSeal(**COMMON_PARAMS, **MANUAL_PARAMS)

    # Compare stiffness coefficients (should be within 5%)
    assert_allclose(holepattern_gas.kxx[0], holepattern_man.kxx[0], rtol=0.05)
    assert_allclose(holepattern_gas.cxx[0], holepattern_man.cxx[0], rtol=0.05)
    assert_allclose(
        holepattern_gas.seal_leakage[0], holepattern_man.seal_leakage[0], rtol=0.01
    )


@pytest.mark.skipif(not REFPROP_AVAILABLE, reason="Requires REFPROP")
def test_holepattern_refprop_backend():
    """Test HolePatternSeal with REFPROP backend explicitly.

    This test only runs when REFPROP is available.
    """
    # Ensure we're using REFPROP
    original_eos = ccp.config.EOS
    try:
        ccp.config.EOS = "REFPROP"

        holepattern = HolePatternSeal(**COMMON_PARAMS, gas_composition=GAS_COMPOSITION)

        # REFPROP expected values (from testing)
        assert_allclose(holepattern.kxx[0], 606104.82, rtol=1e-3)
        assert_allclose(holepattern.cxx[0], 299.71, rtol=1e-3)
        assert_allclose(holepattern.seal_leakage[0], 0.633606, rtol=1e-3)

        # Check derived properties
        assert_allclose(holepattern.molar, 28.9586, rtol=1e-3)
        assert_allclose(holepattern.gamma, 1.4321, rtol=1e-3)

    finally:
        ccp.config.EOS = original_eos


def test_holepattern_heos_backend():
    """Test HolePatternSeal with HEOS backend explicitly.

    This test should always work.
    """
    original_eos = ccp.config.EOS
    try:
        ccp.config.EOS = "HEOS"

        holepattern = HolePatternSeal(**COMMON_PARAMS, gas_composition=GAS_COMPOSITION)

        # HEOS expected values (from testing)
        # Note: HEOS values differ slightly from REFPROP (<0.1%)
        assert_allclose(holepattern.kxx[0], 605751.93, rtol=1e-2)
        assert_allclose(holepattern.cxx[0], 299.76, rtol=1e-2)
        assert_allclose(holepattern.seal_leakage[0], 0.633636, rtol=1e-3)

        # Check derived properties
        assert_allclose(holepattern.molar, 28.9586, rtol=1e-3)
        # HEOS gamma calculation should be similar to REFPROP
        assert_allclose(holepattern.gamma, 1.4321, rtol=1e-2)

    finally:
        ccp.config.EOS = original_eos


@pytest.mark.skipif(not REFPROP_AVAILABLE, reason="Requires REFPROP")
def test_holepattern_refprop_vs_heos_consistency():
    """Compare REFPROP and HEOS results for the same configuration.

    Results should be within 1% of each other.
    """
    original_eos = ccp.config.EOS

    try:
        # Test with REFPROP
        ccp.config.EOS = "REFPROP"
        holepattern_refprop = HolePatternSeal(
            **COMMON_PARAMS, gas_composition=GAS_COMPOSITION
        )

        # Test with HEOS
        ccp.config.EOS = "HEOS"
        holepattern_heos = HolePatternSeal(
            **COMMON_PARAMS, gas_composition=GAS_COMPOSITION
        )

        # Compare coefficients (should be within 1%)
        assert_allclose(
            holepattern_refprop.kxx[0],
            holepattern_heos.kxx[0],
            rtol=0.01,
            err_msg="kxx differs by >1% between REFPROP and HEOS",
        )
        assert_allclose(
            holepattern_refprop.cxx[0],
            holepattern_heos.cxx[0],
            rtol=0.01,
            err_msg="cxx differs by >1% between REFPROP and HEOS",
        )
        assert_allclose(
            holepattern_refprop.seal_leakage[0],
            holepattern_heos.seal_leakage[0],
            rtol=0.001,
            err_msg="seal_leakage differs by >0.1% between REFPROP and HEOS",
        )

        # Derived properties should match closely
        assert_allclose(holepattern_refprop.molar, holepattern_heos.molar, rtol=1e-4)
        assert_allclose(holepattern_refprop.gamma, holepattern_heos.gamma, rtol=1e-2)

    finally:
        ccp.config.EOS = original_eos


def test_holepattern_without_gas_params_raises():
    """Test that creating HolePatternSeal without gas params raises error."""
    with pytest.raises(TypeError):
        # Should fail because neither gas_composition nor manual params provided
        # Raises TypeError: unsupported operand type(s) for /: 'float' and 'NoneType'
        HolePatternSeal(**COMMON_PARAMS)


def test_holepattern_invalid_gas_composition():
    """Test that invalid gas composition raises appropriate error."""
    invalid_gas = {"InvalidFluid123": 1.0}

    with pytest.raises((ValueError, KeyError)):
        HolePatternSeal(**COMMON_PARAMS, gas_composition=invalid_gas)


def test_holepattern_multiple_frequencies():
    """Test HolePatternSeal with multiple frequency points."""
    params = COMMON_PARAMS.copy()
    params["frequency"] = Q_([3000, 5000, 7000], "RPM")

    holepattern = HolePatternSeal(**params, **MANUAL_PARAMS)

    # Should have 3 values for each coefficient
    assert len(holepattern.kxx) == 3
    assert len(holepattern.cxx) == 3
    assert len(holepattern.seal_leakage) == 3

    # All should be positive
    assert all(np.array(holepattern.kxx) > 0)
    assert all(np.array(holepattern.cxx) > 0)
    assert all(np.array(holepattern.seal_leakage) > 0)


@pytest.fixture
def holepattern():
    """Original fixture for backward compatibility."""
    seal = HolePatternSeal(
        n=0,
        frequency=Q_([5000], "RPM"),
        length=0.0254,
        shaft_radius=0.0751,
        radial_clearance=0.0004,
        roughness=0.00198,
        cell_length=0.001,
        cell_width=0.001,
        cell_depth=0.00229,
        inlet_pressure=1830000.0,
        outlet_pressure=823500.0,
        inlet_temperature=300.0,
        b_suther=1.458e-6,
        s_suther=110.4,
        molar=29.0,
        gamma=1.4,
        preswirl=1.0,
        entr_coef=0.1,
        exit_coef=0.5,
    )
    return seal


def test_holepattern_coefficients(holepattern):
    """Original test - kept for backward compatibility."""
    assert_allclose(holepattern.kxx, 586228.88958017, rtol=1e-4)
    assert_allclose(holepattern.kxy, 159741.17580073, rtol=1e-4)
    assert_allclose(holepattern.kyx, -159741.17580073, rtol=1e-4)
    assert_allclose(holepattern.kyy, 586228.88958017, rtol=1e-4)
    assert_allclose(holepattern.cxx, 294.42942927, rtol=1e-4)
    assert_allclose(holepattern.cxy, -27.16217997, rtol=1e-4)
    assert_allclose(holepattern.cyx, 27.16217997, rtol=1e-4)
    assert_allclose(holepattern.cyy, 294.42942927, rtol=1e-4)
    assert_allclose(holepattern.seal_leakage, 0.6313559209954082, rtol=1e-4)



import numpy as np
import math
import time
from ross.bearing_seal_element import BearingElement
from ross.units import Q_, check_units
from ross.bearings.lubricants import lubricants_dict
from prettytable import PrettyTable


class SqueezeFilmDamper(BearingElement):
    """
    Squeeze Film Damper (SFD) element in ROSS standard format.
    Computes damping (co), stiffness (ko), maximum pressure (p_max)
    and pressure angle (theta_m) based on classical short-bearing theory.

    Parameters
    ----------
    Bearing Geometry
    ^^^^^^^^^^^^^^^^
    Describes the geometric characteristics.
    n : int
        Node in which the bearing will be located.
    frequency : list, pint.Quantity
        Array with the frequencies (rad/s).
    axial_length : float, pint.Quantity
        Bearing length. Default unit is meter.
    journal_radius : float
        Rotor radius. The unit is meter.
    radial_clearance : float
        Radial clearence between rotor and bearing. The unit is meter.
    eccentricity_ratio : float
        Normal cases, utilize the eccentricity radio < 0.4, because this
        gap has better propieties. It's a dimensioneless parametre.
    lubricant : str or dict
        Lubricant type. Can be:
        - 'ISOVG32'
        - 'ISOVG46'
        - 'ISOVG68'
        Or a dictionary with lubricant properties.
    geometry : str
        Geometry type. Can be:
        - 'groove': SFD with groove and no end seals
        - 'end_seals': SFD with end seals and no groove
        - 'groove-end_seals': SFD with both groove and end seals
        Default is 'groove'.
    cavitation : boolean
        It can be true or false, depending on configuration of the flow type.
        Default is True.
    tag : str, optional
        Tag for the element.
    scale_factor : float, optional
        Scale factor for the bearing. Default is 1.0.

    Returns
    -------
    Bearing Elements

    Attributes
    ----------
    cxx : ndarray
        Damping coefficient in N*s/m. Shape: (n_freq,).
    kxx : ndarray
        Stiffness coefficient in N/m. Shape: (n_freq,).
    theta : ndarray
        Pressure angle in radians. Shape: (n_freq,).
    p_max : ndarray
        Maximum pressure in Pa. Shape: (n_freq,).

    Example
    -------
    >>> import ross as rs
    >>> Q_ = rs.units.Q_
    >>> SFD = rs.SqueezeFilmDamper(
    ...     n=0,
    ...     frequency=Q_([18600, 20000, 22000], "rpm"),
    ...     axial_length=Q_(0.9, "inches"),
    ...     journal_radius=Q_(2.55, "inches"),
    ...     radial_clearance=Q_(0.003, "inches"),
    ...     eccentricity_ratio=0.5,
    ...     lubricant="ISOVG32",
    ...     geometry="groove",
    ...     cavitation=True,
    ... )
    """

    @check_units
    def __init__(
        self,
        n,
        frequency,
        axial_length,
        journal_radius,
        radial_clearance,
        eccentricity_ratio,
        lubricant,
        geometry="groove",
        cavitation=True,
        tag=None,
        scale_factor=1.0,
    ):
        self.axial_length = axial_length
        self.journal_radius = journal_radius
        self.radial_clearance = radial_clearance
        self.eccentricity_ratio = eccentricity_ratio
        self.frequency = frequency
        self.lubricant = lubricants_dict[lubricant]["liquid_viscosity1"]
        self.cavitation = cavitation
        self.geometry = geometry

        # Start timing
        self.initial_time = time.time()

        # Get number of frequencies
        n_freq = np.shape(self.frequency)[0]

        # Initialize arrays for storing results at each frequency
        kxx = np.zeros(n_freq)
        cxx = np.zeros(n_freq)
        theta_array = np.zeros(n_freq)
        p_max_array = np.zeros(n_freq)

        # Calculate coefficients for each frequency
        for i in range(n_freq):
            freq = self.frequency[i]

            # Calculate coefficients based on geometry
            if self.geometry == "groove":
                co, ko, theta, p_max = self.calculate_coeficients_with_groove(freq)
            elif self.geometry == "end_seals":
                co, ko, theta, p_max = self.calculate_coeficients_with_end_seals(freq)
            elif self.geometry == "groove-end_seals":
                co, ko, theta, p_max = (
                    self.calculate_coeficientes_with_groove_and_end_seals(freq)
                )
            else:
                raise ValueError(
                    f"Invalid geometry type: {geometry}. "
                    "Must be 'groove', 'end_seals', or 'groove-end_seals'"
                )

            # Store results for this frequency
            kxx[i] = ko
            cxx[i] = co
            theta_array[i] = theta
            p_max_array[i] = p_max

        # Store calculated values as instance attributes
        self.theta = theta_array
        self.p_max = p_max_array

        # End timing
        self.final_time = time.time()

        super().__init__(
            n=n,
            frequency=frequency,
            kxx=kxx,
            cxx=cxx,
            tag=tag,
            scale_factor=scale_factor,
        )

    def calculate_coeficients_with_end_seals(self, freq):
        """Calculate coefficients for a sealed SFD without a groove.

        Parameters
        ----------
        freq : float
            Operating frequency in rad/s.

        Returns
        -------
        co : float
            Damping coefficient.
        ko : float
            Stiffness coefficient.
        theta : float
            Pressure angle in radians.
        p_max : float
            Maximum pressure.
        """
        co = (
            12.0
            * np.pi
            * self.axial_length
            * (self.journal_radius / self.radial_clearance) ** 3
            * self.lubricant
        )
        co /= (2.0 + self.eccentricity_ratio**2) * np.sqrt(
            1.0 - self.eccentricity_ratio**2
        )

        ko = (
            24.0
            * self.lubricant
            * self.axial_length
            * (self.journal_radius / self.radial_clearance) ** 3
            * self.eccentricity_ratio
            * freq
        )
        ko /= (2.0 + self.eccentricity_ratio**2) * (1.0 - self.eccentricity_ratio**2)

        theta_m = -80.45 * self.eccentricity_ratio + 268.98
        theta = math.radians(theta_m)

        p_max_num = (
            2.0
            * self.eccentricity_ratio
            * (2.0 + self.eccentricity_ratio * np.cos(theta))
            * np.sin(theta)
        )
        p_max_den = (2.0 + self.eccentricity_ratio**2) * (
            1.0 + self.eccentricity_ratio * np.cos(theta)
        ) ** 2
        p_max = (
            -p_max_num
            / p_max_den
            * 6.0
            * self.lubricant
            * freq
            * (self.journal_radius / self.radial_clearance) ** 2
        )

        if self.cavitation:
            ko = 0.0
        else:
            co = 2.0 * co
            ko = 0.0

        return co, ko, theta, p_max

    def calculate_coeficients_with_groove(self, freq):
        """Calculate coefficients for an SFD with a groove and no end seals.

        Parameters
        ----------
        freq : float
            Operating frequency in rad/s.

        Returns
        -------
        co : float
            Damping coefficient.
        ko : float
            Stiffness coefficient.
        theta : float
            Pressure angle in radians.
        p_max : float
            Maximum pressure.
        """
        if self.cavitation:
            co = (
                self.lubricant
                * (self.axial_length**3)
                * self.journal_radius
                / (2.0 * self.radial_clearance**3)
            )
            co *= np.pi / ((1.0 - self.eccentricity_ratio**2) ** (3.0 / 2.0))
            co /= 4

            ko = (
                2.0
                * self.lubricant
                * freq
                * self.journal_radius
                * (self.axial_length / self.radial_clearance) ** 3
                * self.eccentricity_ratio
            )
            ko /= (1.0 - self.eccentricity_ratio**2) ** 2
            ko /= 4

        theta_m = (
            270.443
            - 191.831 * self.eccentricity_ratio
            + 218.223 * self.eccentricity_ratio**2
            - 114.803 * self.eccentricity_ratio**3
        )
        theta = math.radians(theta_m)

        p_max = (
            -1.5
            * (self.axial_length / self.radial_clearance) ** 2
            * self.lubricant
            * freq
            * self.eccentricity_ratio
            * np.sin(theta)
        )
        p_max /= (1.0 + self.eccentricity_ratio * np.cos(theta)) ** 3
        p_max /= 2

        if not self.cavitation:
            ko = 0.0
            co = (
                self.lubricant
                * (self.axial_length / self.radial_clearance) ** 3
                * self.journal_radius
                * np.pi
            )
            co /= (1.0 - self.eccentricity_ratio**2) ** (3.0 / 2.0)

        return co, ko, theta, p_max

    def calculate_coeficientes_with_groove_and_end_seals(self, freq):
        """Calculate coefficients for an SFD with both a groove and end seals.

        Parameters
        ----------
        freq : float
            Operating frequency in rad/s.

        Returns
        -------
        co : float
            Damping coefficient.
        ko : float
            Stiffness coefficient.
        theta : float
            Pressure angle in radians.
        p_max : float
            Maximum pressure.
        """
        if self.cavitation:
            co = (
                self.lubricant
                * (self.axial_length**3)
                * self.journal_radius
                / (2.0 * self.radial_clearance**3)
            )
            co *= np.pi / ((1.0 - self.eccentricity_ratio**2) ** (3.0 / 2.0))

            ko = (
                2.0
                * self.lubricant
                * freq
                * self.journal_radius
                * (self.axial_length / self.radial_clearance) ** 3
                * self.eccentricity_ratio
            )
            ko /= (1.0 - self.eccentricity_ratio**2) ** 2

        theta_m = (
            270.443
            - 191.831 * self.eccentricity_ratio
            + 218.223 * self.eccentricity_ratio**2
            - 114.803 * self.eccentricity_ratio**3
        )
        theta = math.radians(theta_m)

        p_max = (
            -1.5
            * (self.axial_length / self.radial_clearance) ** 2
            * self.lubricant
            * freq
            * self.eccentricity_ratio
            * np.sin(theta)
        )
        p_max /= (1.0 + self.eccentricity_ratio * np.cos(theta)) ** 3

        if not self.cavitation:
            ko = 0.0
            co = (
                self.lubricant
                * (self.axial_length / self.radial_clearance) ** 3
                * self.journal_radius
                * np.pi
            )
            co /= (1.0 - self.eccentricity_ratio**2) ** (3.0 / 2.0)

        return co, ko, theta, p_max

    def show_results(self):
        """Display Squeeze Film Damper calculation results in a formatted table.

        This method prints the main results from the SFD analysis using PrettyTable,
        including geometric parameters, operating conditions, and calculated coefficients
        for each frequency.

        Parameters
        ----------
        None
            This method uses the damper parameters and results stored as
            instance attributes.

        Returns
        -------
        None
            Results are printed to the console in a formatted table.
        """
        if self.frequency.size == 1:
            self._print_single_frequency_results(0)
        else:
            for i in range(self.frequency.size):
                self._print_single_frequency_results(i)

    def _print_single_frequency_results(self, freq_index):
        """Print results for a single frequency."""
        freq = self.frequency[freq_index]

        # Define a fixed width for all columns
        column_width = 20

        table = PrettyTable()
        table.field_names = ["Parameter", "Value", "Unit"]

        for field in table.field_names:
            table.max_width[field] = column_width
            table.min_width[field] = column_width

        # Set column alignment
        table.align["Parameter"] = "l"
        table.align["Value"] = "r"
        table.align["Unit"] = "c"

        # Operating conditions
        table.add_row(["Operating Speed", f"{freq * 30 / np.pi:12.1f}", "RPM"])
        table.add_row(["Geometry Type", f"{self.geometry:>12}", "-"])
        table.add_row(["Cavitation", f"{str(self.cavitation):>12}", "-"])

        # Geometric parameters
        table.add_row(["Axial Length", f"{self.axial_length:12.6f}", "m"])
        table.add_row(["Journal Radius", f"{self.journal_radius:12.6f}", "m"])
        table.add_row(["Radial Clearance", f"{self.radial_clearance:12.6e}", "m"])
        table.add_row(["Eccentricity Ratio", f"{self.eccentricity_ratio:12.4f}", "-"])

        # Lubricant viscosity
        table.add_row(["Lubricant Viscosity", f"{self.lubricant:12.4e}", "Pa*s"])

        # Display stored coefficients (already calculated during __init__)
        table.add_row(["Damping Coefficient", f"{self.cxx[freq_index]:12.4e}", "N*s/m"])
        table.add_row(["Stiffness Coefficient", f"{self.kxx[freq_index]:12.4e}", "N/m"])

        # Display stored pressure angle and max pressure for this frequency
        theta_val = self.theta[freq_index]
        p_max_val = self.p_max[freq_index]

        table.add_row(["Pressure Angle", f"{np.degrees(theta_val):12.2f}", "°"])
        table.add_row(["Pressure Angle", f"{theta_val:12.4f}", "rad"])
        table.add_row(["Maximum Pressure", f"{p_max_val:12.4e}", "Pa"])

        table_str = table.get_string()
        final_width = len(table_str.split("\n")[0])

        print("\n" + "=" * final_width)
        print(
            f"SQUEEZE FILM DAMPER RESULTS - {freq * 30 / np.pi:.1f} RPM".center(
                final_width
            )
        )
        print("=" * final_width)
        print(table)
        print("=" * final_width)

    def show_coefficients_comparison(self):
        """Display SFD coefficients comparison table across frequencies.

        This method creates and displays a formatted table comparing damping
        and stiffness coefficients across different frequencies.

        Parameters
        ----------
        None
            This method uses the frequency array and coefficients stored as
            instance attributes.

        Returns
        -------
        None
            Results are printed to the console in a formatted table.
        """
        freq_rpm = np.atleast_1d(self.frequency).astype(float) * 30.0 / np.pi

        table = PrettyTable()

        headers = [
            "Frequency [RPM]",
            "cxx [N*s/m]",
            "kxx [N/m]",
            "Pressure [Pa]",
            "Angle [°]",
        ]

        table.field_names = headers

        for i in range(len(freq_rpm)):
            theta_val = self.theta[i]
            p_max_val = self.p_max[i]

            row = [
                f"{freq_rpm[i]:.1f}",
                f"{self.cxx[i]:.4e}",
                f"{self.kxx[i]:.4e}",
                f"{p_max_val:.4e}",
                f"{np.degrees(theta_val):.2f}",
            ]

            table.add_row(row)

        # Table width
        desired_width = 20

        table.max_width = desired_width
        table.min_width = desired_width

        table_str = table.get_string()
        table_lines = table_str.split("\n")
        actual_width = len(table_lines[0])

        print("\n" + "=" * actual_width)
        print("SFD COEFFICIENTS COMPARISON TABLE".center(actual_width))
        print("=" * actual_width)
        print(table)
        print("=" * actual_width)

    def show_execution_time(self):
        """Display the simulation execution time.

        This method calculates and displays the total time spent during the
        complete bearing analysis execution, including all frequency calculations.

        Parameters
        ----------
        None
            This method uses the initial_time and final_time attributes
            stored during the simulation execution.

        Returns
        -------
        float
            Total simulation time in seconds. Returns None if simulation
            hasn't been executed yet.
        """
        if hasattr(self, "initial_time") and hasattr(self, "final_time"):
            total_time = self.final_time - self.initial_time
            print(f"Execution time: {total_time:.6f} seconds")
        else:
            print("Simulation hasn't been executed yet.")



from abc import ABC
from pathlib import Path

import numpy as np
import pandas as pd

import ross as rs
from ross.units import Q_, check_units

__all__ = ["Crack"]


class Crack(ABC):
    """Model a crack on a given shaft element of a rotor system.
    The Gasch and Mayes models are based on Linear Fracture Mechanics,
    while the Flex models are based on an equivalent beam model from 3D finite element computations.

    Contains transversal crack models :cite:`gasch1993survey` and :cite:`mayes1984analysis`.
    The reference coordenate system is:
        - x-axis and y-axis in the sensors' planes;
        - z-axis throught the shaft center.

    Parameters
    ----------
    rotor : ross.Rotor
        Rotor object.
    n : float
        Number of shaft element where crack is located.
    depth_ratio : float
        Crack depth ratio related to the diameter of the crack container element.
        A depth value of 0.1 is equal to 10%, 0.2 equal to 20%, and so on.
    crack_model : string, optional
        Name of the chosen crack model. The available types are: "Mayes", "Gasch",
        "Flex Open" and "Flex Breathing".
        Default is "Mayes".
    cross_divisions: float, optional
        Number of square divisions into which the cross-section of the cracked element
        will be divided in the analysis conducted for the Flex Breathing model.

    Returns
    -------
    A crack object.

    Attributes
    ----------
    shaft_elem : ross.ShaftElement
        A 6 degrees of freedom shaft element object where crack is located.
    K_elem : np.ndarray
        Stiffness matrix of the shaft element without crack.
    Ko : np.ndarray
        Stiffness of the shaft with the crack closed (equivalent to the shaft without crack).
    Kc : np.ndarray
        Stiffness of the shaft including compliance coefficients according to the crack depth.
    forces : np.ndarray
        Force matrix due to crack. Each row corresponds to a dof and each column to a time.

    References
    ----------
    .. bibliography::
        :filter: docname in docnames

    Examples
    --------
    >>> rotor = rs.rotor_example_with_damping()
    >>> fault = Crack(rotor, n=18, depth_ratio=0.2, crack_model="Gasch")
    >>> fault.shaft_elem
    ShaftElement(L=0.03, idl=0.0, idr=0.0, odl=0.019,  odr=0.019, material='Steel', n=18)
    """

    @check_units
    def __init__(
        self,
        rotor,
        n,
        depth_ratio,
        crack_model="Mayes",
        cross_divisions=None,
    ):
        self.rotor = rotor
        self.cross_divisions = cross_divisions
        self.crack_model = crack_model

        self.validate_depth_ratio(depth_ratio, crack_model)
        self.depth_ratio = depth_ratio

        if crack_model is None or crack_model == "Mayes":
            self._crack_model = self.mayes
        elif crack_model == "Gasch":
            self._crack_model = self.gasch
        elif crack_model == "Flex Open":
            self._crack_model = self.flex_open
        elif crack_model == "Flex Breathing":
            self._crack_model = self.flex_breathing
            if cross_divisions is None:
                raise ValueError(
                    "The 'cross_divisions' argument must be provided for the Flex Breathing model."
                )
        else:
            raise Exception("Check the crack model!")

        # Shaft element with crack
        self.shaft_elem = [elm for elm in rotor.shaft_elements if elm.n == n][0]

        self.dofs = list(self.shaft_elem.dof_global_index.values())

        self.K_elem = self.shaft_elem.K()

        if self._crack_model == self.mayes or self._crack_model == self.gasch:
            dir_path = Path(__file__).parents[0] / "data/PAPADOPOULOS.csv"
            self.coefficient_data = pd.read_csv(dir_path)

            L = self.shaft_elem.L
            E = self.shaft_elem.material.E
            Ie = self.shaft_elem.Ie
            phi = self.shaft_elem.phi

            co1 = L**3 * (1 + phi / 4) / 3
            co2 = L**2 / 2
            co3 = L

            # fmt: off
            Co = np.array([
                [co1,   0,     0, co2],
                [  0,  co1, -co2,   0],
                [  0, -co2,  co3,   0],
                [co2,    0,    0, co3],
            ]) / (E * Ie)
            # fmt: on

            if self.depth_ratio == 0:
                Cc = Co
            else:
                c44 = self._get_coefficient("c44")
                c55 = self._get_coefficient("c55")
                c45 = self._get_coefficient("c45")

                Cc = Co + np.array(
                    [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, c55, c45], [0, 0, c45, c44]]
                )

            self.Ko = np.linalg.pinv(Co)
            self.Kc = np.linalg.pinv(Cc)

    @staticmethod
    def validate_depth_ratio(depth_ratio, crack_model):
        """Validate the maximum allowed crack depth ratio for each crack model.
        Each crack model has a specific upper limit for the relative depth
        of the crack. This function raises a ValueError if the
        input value exceeds the supported range.

        Parameters
        ----------
        depth_ratio : float
            Crack depth ratio related to the diameter of the crack container element.
            A depth value of 0.1 is equal to 10%, 0.2 equal to 20%, and so on.
        crack_model : string, optional
            Name of the chosen crack model. The available types are: "Mayes", "Gasch",
            "Flex Open" and "Flex Breathing".
            Default is "Mayes".

        Raises
        ------
        ValueError
            If depth_ratio exceeds the supported limit for the given model.
        """

        limits = {
            "Mayes": 0.5,
            "Gasch": 0.5,
            "Flex Open": 0.6,
            "Flex Breathing": 0.6,
        }

        max_allowed = limits.get(crack_model)
        if max_allowed is not None and depth_ratio > max_allowed:
            raise ValueError(
                f"""
                For cracks deeper than {max_allowed * 100:.0f}% of diameter, this approach has a singularity and cannot be used.
                """
            )

    def _get_coefficient(self, coeff):
        """Get terms of the compliance matrix.

        Parameters
        -----------
        coeff : string
            Name of the coefficient according to the corresponding direction.

        Returns
        -------
        c : np.ndarray
            Compliance coefficient according to the crack depth.
        """

        Poisson = self.shaft_elem.material.Poisson
        E = self.shaft_elem.material.E
        radius = self.shaft_elem.odl / 2

        c = np.array(pd.eval(self.coefficient_data[coeff]))
        ind = np.where(c[:, 1] >= self.depth_ratio * 2)[0]

        c = c[ind[0], 0] * (1 - Poisson**2) / (E * radius**3)

        return c

    def _compute_crack_stiffness_gasch_mayes(self, Kmodel):
        """Compute stiffness matrix of the shaft element with crack in inertial coordinates
        for the Gasch and Mayes models.

        Parameters
        ----------
        ap : float
            Angular position of the element.
        Kmodel: np.ndarray
            Stiffness matrix of the shaft element with crack in rotating coordinates.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix of the cracked element.
        """

        L = self.shaft_elem.L
        Toxy = np.array([[-1, 0], [-L, -1], [1, 0], [0, 1]])
        kxy = np.array([[Kmodel[0, 0], self.Ko[0, 3]], [self.Ko[3, 0], self.Ko[3, 3]]])
        Koxy = Toxy @ kxy @ Toxy.T

        Toyz = np.array([[-1, 0], [L, -1], [1, 0], [0, 1]])
        kyz = np.array([[Kmodel[1, 1], self.Ko[1, 2]], [self.Ko[2, 1], self.Ko[2, 2]]])
        Koyz = Toyz @ kyz @ Toyz.T

        # fmt: off
        K = np.array([
            [Koxy[0,0],         0,   0,         0, Koxy[0,1],   0, Koxy[0,2],         0,   0,         0, Koxy[0,3],   0],
            [        0, Koyz[0,0],   0, Koyz[0,1],         0,   0,         0, Koyz[0,2],   0, Koyz[0,3],         0,   0],
            [        0,         0,   0,         0,         0,   0,         0,         0,   0,         0,         0,   0],
            [        0, Koyz[1,0],   0, Koyz[1,1],         0,   0,         0, Koyz[1,2],   0, Koyz[1,3],         0,   0],
            [Koxy[1,0],         0,   0,         0, Koxy[1,1],   0, Koxy[1,2],         0,   0,         0, Koxy[1,3],   0],
            [        0,         0,   0,         0,         0,   0,         0,         0,   0,         0,         0,   0],
            [Koxy[2,0],         0,   0,         0, Koxy[2,1],   0, Koxy[2,2],         0,   0,         0, Koxy[2,3],   0],
            [        0, Koyz[2,0],   0, Koyz[2,1],         0,   0,         0, Koyz[2,2],   0, Koyz[2,3],         0,   0],
            [        0,         0,   0,         0,         0,   0,         0,         0,   0,         0,         0,   0],
            [        0, Koyz[3,0],   0, Koyz[3,1],         0,   0,         0, Koyz[3,2],   0, Koyz[3,3],         0,   0],
            [Koxy[3,0],         0,   0,         0, Koxy[3,1],   0, Koxy[3,2],         0,   0,         0, Koxy[3,3],   0],
            [        0,         0,   0,         0,         0,   0,         0,         0,   0,         0,         0,   0]
        ])
        # fmt: on

        return K

    def gasch(self, ap):
        """Stiffness matrix of the shaft element with crack in rotating coordinates
        according to the breathing model of Gasch.
        This model is based on Linear Fracture Mechanics.

        Paramenters
        -----------
        ap : float
            Angular position of the shaft.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix of the cracked element.
        """

        # Gasch
        ko = self.Ko[0, 0]
        kcx = self.Kc[0, 0]
        kcz = self.Kc[1, 1]

        kme = (ko + kcx) / 2
        kmn = (ko + kcz) / 2
        kde = (ko - kcx) / 2
        kdn = (ko - kcz) / 2

        size = 18
        cosine_sum = np.sum(
            [(-1) ** i * np.cos((2 * i + 1) * ap) / (2 * i + 1) for i in range(size)]
        )

        ke = kme + (4 / np.pi) * kde * cosine_sum
        kn = kmn + (4 / np.pi) * kdn * cosine_sum

        T_matrix = np.array(
            [
                [np.cos(ap), np.sin(ap)],
                [-np.sin(ap), np.cos(ap)],
            ]
        )

        K = T_matrix.T @ np.array([[ke, 0], [0, kn]]) @ T_matrix

        return self._compute_crack_stiffness_gasch_mayes(K)

    def mayes(self, ap):
        """Stiffness matrix of the shaft element with crack in rotating coordinates
        according to the breathing model of Mayes.
        This model is based on Linear Fracture Mechanics.

        Paramenters
        -----------
        ap : float
            Angular position of the shaft.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix of the cracked element.
        """

        # Mayes
        ko = self.Ko[0, 0]
        kcx = self.Kc[0, 0]
        kcz = self.Kc[1, 1]

        ke = 0.5 * (ko + kcx) + 0.5 * (ko - kcx) * np.cos(ap)
        kn = 0.5 * (ko + kcz) + 0.5 * (ko - kcz) * np.cos(ap)

        T_matrix = np.array(
            [
                [np.cos(ap), np.sin(ap)],
                [-np.sin(ap), np.cos(ap)],
            ]
        )

        K = T_matrix.T @ np.array([[ke, 0], [0, kn]]) @ T_matrix

        return self._compute_crack_stiffness_gasch_mayes(K)

    def _compute_crack_stiffness_flex(self, Lce, at, IXX, IYY, IXY):
        """Compute stiffness matrix of the shaft element with crack in inertial coordinates
        for the Flex Open and Flex Breathing models.

        Parameters
        ----------
        Lce: float
            Equivalent cracked element length.
        at: float
            Resistant area.
        IXX: float
            Moments of inertia about the x axis.
        IYY: float
            Moment of inertia about the y axis.
        IXY: float
            Product of inertia.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix of the cracked element.
        """
        E = self.shaft_elem.material.E
        Poisson = self.shaft_elem.material.Poisson
        G_s = self.shaft_elem.material.G_s
        kcc = (6 * (1 + Poisson)) / (7 + 6 * Poisson)

        fiiY = 12 * E * IXX / (G_s * kcc * at * Lce**2)
        fiiX = 12 * E * IYY / (G_s * kcc * at * Lce**2)
        fiiXY = 12 * E * IYY / (G_s * kcc * at * Lce**2)

        aF = 12 * IYY * E / ((1 + fiiX) * Lce**3)
        bF = 12 * IXX * E / ((1 + fiiY) * Lce**3)
        cF = 6 * IYY * E / ((1 + fiiX) * Lce**2)
        dF = 6 * IXX * E / ((1 + fiiY) * Lce**2)
        eF = (4 + fiiX) * IYY * E / ((1 + fiiX) * Lce)
        fF = (2 - fiiX) * IYY * E / ((1 + fiiX) * Lce)
        gF = (2 - fiiY) * IXX * E / ((1 + fiiY) * Lce)
        hF = (4 + fiiY) * IXX * E / ((1 + fiiY) * Lce)
        pF = 12 * IXY * E / ((1 + fiiXY) * Lce**3)
        qF = 6 * IXY * E / ((1 + fiiXY) * Lce**2)
        rF = (4 + fiiXY) * IXY * E / ((1 + fiiXY) * Lce)
        sF = (2 - fiiXY) * IXY * E / ((1 + fiiXY) * Lce)
        tF = 0
        uF = 0
        wF = 0
        kF = 0
        mF = 0
        nF = 0
        iF = 0
        jF = 0

        K = np.array(
            [
                [bF, -pF, kF, -qF, -dF, nF, -bF, pF, -kF, -qF, -dF, -mF],
                [-pF, aF, wF, cF, qF, mF, pF, -aF, -wF, cF, qF, -mF],
                [kF, tF, wF, iF, 0, jF, -kF, -tF, -wF, -iF, 0, -jF],
                [-qF, cF, iF, eF, rF, 0, qF, -cF, -iF, fF, sF, 0],
                [-dF, qF, jF, rF, hF, 0, dF, -qF, -jF, sF, gF, 0],
                [nF, 0, mF, 0, uF, 0, -nF, 0, -mF, 0, -uF, 0],
                [-bF, pF, -kF, qF, dF, -nF, bF, -pF, kF, qF, dF, nF],
                [pF, -aF, -wF, -cF, -qF, -mF, -pF, aF, wF, -cF, -qF, mF],
                [-kF, -tF, -wF, -iF, 0, -jF, 0, tF, wF, mF, 0, nF],
                [-qF, cF, -iF, fF, sF, 0, qF, -cF, iF, eF, rF, 0],
                [-dF, qF, -jF, sF, gF, 0, dF, -qF, 0, rF, hF, 0],
                [-mF, 0, -mF, 0, -uF, 0, 0, 0, mF, 0, uF, 0],
            ]
        )

        return K

    def flex_open(self, ap):
        """Prepare crack geometry parameters for the Flex Open model.
        Assuming the crack remains fully open during rotation, the resistant area
        and inertia moments are determined based on the crack depth and angular
        position. These are then used to compute the element stiffness.

        Paramenters
        -----------
        ap : float
            Angular position of the shaft.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix of the cracked element.
        """

        # Flex Open
        radius = self.shaft_elem.odl / 2
        depth_ratio = self.depth_ratio
        Lce = (
            (
                -33.3333 * depth_ratio**5
                + 77.5641 * depth_ratio**4
                - 49.5542 * depth_ratio**3
                + 9.0485 * depth_ratio**2
                + 1.2415 * depth_ratio
                - 0.0024
            )
            * 2
            * radius
        )

        mi = 2 * depth_ratio
        gama = (mi * (2 - mi)) ** (1 / 2)
        at = radius**2 * (
            np.pi - np.arccos(1 - mi) + (1 - mi) * (mi * (2 - mi)) ** (1 / 2)
        )
        ee = 2 * radius**3 / (3 * at) * (mi * (2 - mi)) ** (3 / 2)

        Ix = np.pi * radius**4 / 8 + radius**4 / 4 * (
            (1 - mi) * (2 * mi**2 - 4 * mi + 1) * gama + np.arcsin(1 - mi)
        )
        Iy = np.pi * radius**4 / 4 - radius**4 / 12 * (
            (1 - mi) * (2 * mi**2 - 4 * mi - 3) * gama + 3 * np.arcsin(gama)
        )
        Ixb = Ix - at * ee**2
        Iyb = Iy

        IXX = (Ixb + Iyb) / 2 + (Ixb - Iyb) / 2 * np.cos(2 * ap)
        IYY = (Ixb + Iyb) / 2 - (Ixb - Iyb) / 2 * np.cos(2 * ap)
        IXY = -(Ixb - Iyb) / 2 * np.sin(2 * ap)

        return self._compute_crack_stiffness_flex(Lce, at, IXX, IYY, IXY)

    def flex_breathing(self, ap):
        """Prepare crack geometry parameters for the Flex Breathing model.
        The resistant area and inertia moments are determined iteratively,
        considering which parts of the crack are open or closed under local
        stress. These are then used to compute the element stiffness.

        Paramenters
        -----------
        ap : float
            Angular position of the shaft.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix of the cracked element.
        """

        # Flex Breathing
        E = self.shaft_elem.material.E
        radius = self.shaft_elem.odl / 2
        depth_ratio = self.depth_ratio
        J = (np.pi / 4) * radius**4
        step = radius / self.cross_divisions
        Lce = (
            (
                -33.3333 * depth_ratio**5
                + 77.5641 * depth_ratio**4
                - 49.5542 * depth_ratio**3
                + 9.0485 * depth_ratio**2
                + 1.2415 * depth_ratio
                - 0.0024
            )
            * 2
            * radius
        )

        rotation_indexes = [3, 4, 5, 9, 10, 11]
        rot_resp = self.disp_resp[rotation_indexes]
        tx1, ty1, tx2, ty2 = rot_resp[0], rot_resp[1], rot_resp[2], rot_resp[3]

        Mdx = E * J * (tx2 - tx1) / Lce
        Mdy = E * J * (ty2 - ty1) / Lce
        Mpx, Mpy = 0, 0
        MTx = Mdx + Mpx
        MTy = Mdy + Mpy

        CCi = np.zeros(
            (2 * self.cross_divisions + 1, 2 * self.cross_divisions + 1), dtype=complex
        )
        CC = np.zeros_like(CCi, dtype=complex)
        points_crack_i = np.zeros_like(CCi, dtype=complex)
        A = np.zeros(
            (2 * self.cross_divisions, 2 * self.cross_divisions), dtype=complex
        )
        area = np.zeros_like(A, dtype=complex)
        IXX = J
        IYY = J
        IXY = 1j * 0
        XY = np.zeros_like(A, dtype=complex)

        n_points = 2 * self.cross_divisions + 1
        x = np.linspace(-radius, radius, n_points)
        y = np.linspace(-radius, radius, n_points)
        X, Y = np.meshgrid(x, y)
        Z = X + 1j * Y
        abs_z = np.abs(Z)
        angle_z = np.angle(Z)
        CCi = abs_z * np.exp(1j * angle_z)
        CC = abs_z * np.exp(1j * (angle_z + ap))

        mask_depth = np.imag(CCi) > (radius - depth_ratio * radius)
        points_crack_i = np.where(mask_depth, CCi, 1e3)
        points_crack = np.absolute(points_crack_i) * np.exp(
            1j * (np.angle(points_crack_i) + ap)
        )

        exp_step_angle = np.exp(1j * np.angle((step / 2) + (step / 2) * 1j))
        abs_step_angle = np.absolute((step / 2) + (step / 2) * 1j)

        n1, m1 = n_points - 1, n_points - 1

        step_count = 1
        si = 0
        error = 1e3
        previous_angle = 0j

        while error > 1e-5:
            stress = -(MTx * IYY + MTy * IXY) / (IXX * IYY - IXY) ** 2 * np.imag(XY) + (
                MTy * IXX + MTx * IXY
            ) / (IXX * IYY - IXY) ** 2 * np.real(XY)

            c_center = CC[:n1, :m1]
            c_right = CC[:n1, 1:]
            c_down = CC[1:, :m1]
            points_crack_center = points_crack[:n1, :m1]

            A = np.abs((c_center - c_right) * (c_center - c_down))
            area = c_center + abs_step_angle * exp_step_angle
            mask_out = np.abs(area) >= radius
            mask_crack = (np.abs(points_crack_center - c_center) < 1e-8) & (
                np.abs(stress) >= 0
            )
            A[mask_out | mask_crack] = 0
            area[mask_out] = 0
            at = np.sum(A)

            CG = np.sum(area * A) / at if at != 0 else 0

            XY = area - CG
            XY[mask_out] = 0

            dx = c_center - c_right
            dy = c_center - c_down
            dx3 = np.abs(dx) ** 3
            dy3 = np.abs(dy) ** 3

            Ixx = np.abs(dx * dy3) / 12.0 + A * np.abs(np.imag(XY)) ** 2
            Iyy = np.abs(dx3 * dy) / 12.0 + A * np.abs(np.real(XY)) ** 2
            Ixy = A * np.imag(XY) * np.real(XY)
            mask_crack_2 = (np.abs(points_crack_center - c_center) < 1e-8) & (
                np.abs(stress) > 0
            )
            Ixx[mask_out | mask_crack_2] = 0
            Iyy[mask_out | mask_crack_2] = 0
            Ixy[mask_out | mask_crack_2] = 0

            IXX = np.sum(Ixx)
            IYY = np.sum(Iyy)
            IXY = np.sum(Ixy)

            ang = np.arctan(-IXY / (IXX - IYY) / 2) / 2

            if step_count >= 2:
                error = np.abs(np.abs(previous_angle) - np.abs(ang))

            previous_angle = ang
            step_count += 1
            si += 1
            if si >= 50:
                break

        return self._compute_crack_stiffness_flex(Lce, at, IXX, IYY, IXY)

    def _get_force_in_time(self, step, disp_resp, ang_pos):
        """Calculate the dynamic force related on given time step.

        Paramenters
        -----------
        step : int
            Current time step index.
        disp_resp : np.ndarray
            Displacement response of the system at the current time step.
        ang_pos : float
            Angular position of the shaft at the current time step.

        Returns
        -------
        F : np.ndarray
            Force matrix related to the open crack in the current time step `t[step]`.
        """

        self.disp_resp = disp_resp

        K_crack = self._crack_model(ang_pos)

        F = np.zeros(self.rotor.ndof)
        F[self.dofs] = (self.K_elem - K_crack) @ disp_resp[self.dofs]
        self.forces[:, step] = F

        return F

    def run(self, node, unb_magnitude, unb_phase, speed, t, **kwargs):
        """Run analysis for the system with crack given an unbalance force.

        System time response is simulated considering weight force.

        Parameters
        ----------
        node : list, int
            Node where the unbalance is applied.
        unb_magnitude : list, float
            Unbalance magnitude (kg.m).
        unb_phase : list, float
            Unbalance phase (rad).
        speed : float or array_like, pint.Quantity
            Rotor speed.
        t : array
            Time array.
        **kwargs : optional
            Additional keyword arguments can be passed to define the parameters
            of the Newmark method if it is used (e.g. gamma, beta, tol, ...).
            See `ross.utils.newmark` for more details.
            Other keyword arguments can also be passed to be used in numerical
            integration (e.g. num_modes).
            See `Rotor.integrate_system` for more details.

        Returns
        -------
        results : ross.TimeResponseResults
            For more information on attributes and methods available see:
            :py:class:`ross.TimeResponseResults`
        """

        rotor = self.rotor

        # Unbalance force
        F, ang_pos, _, _ = rotor.unbalance_force_over_time(
            node, unb_magnitude, unb_phase, speed, t, return_all=True
        )

        # Weight
        F += rotor.gravitational_force()[:, np.newaxis]

        self.forces = np.zeros((rotor.ndof, len(t)))

        force_crack = lambda step, **state: self._get_force_in_time(
            step, state.get("disp_resp"), ang_pos[step]
        )

        results = rotor.run_time_response(
            speed=speed,
            F=F.T,
            t=t,
            method="newmark",
            add_to_RHS=force_crack,
            **kwargs,
        )

        return results


def crack_example():
    """Create an example to evaluate the influence of transverse cracks in a
    rotating shaft.

    This function returns time response results of a transversal crack fault.
    The purpose is to make available a simple example so that a doctest can be
    written using it.

    Returns
    -------
    results : ross.TimeResponseResults
        Results for a shaft with crack.

    Examples
    --------
    >>> from ross.faults.crack import crack_example
    >>> from ross.probe import Probe
    >>> results = crack_example()
    Running direct method
    >>> probe1 = Probe(14, 0)
    >>> probe2 = Probe(22, 0)
    >>> fig = results.plot_1d([probe1, probe2])
    """

    rotor = rs.rotor_example_with_damping()

    n1 = rotor.disk_elements[0].n
    n2 = rotor.disk_elements[1].n

    results = rotor.run_crack(
        n=18,
        depth_ratio=0.2,
        node=[n1, n2],
        unbalance_magnitude=[5e-4, 0],
        unbalance_phase=[-np.pi / 2, 0],
        crack_model="Mayes",
        speed=Q_(1200, "RPM"),
        t=np.arange(0, 0.5, 0.0001),
    )

    return results



"""Script to run and save all documentation notebooks.

This is useful to update the notebooks before each release.
"""

import nbformat
from pathlib import Path
from nbclient import NotebookClient
from nbclient.exceptions import CellExecutionError, CellTimeoutError

path = Path.cwd()

for notebook_filename in path.rglob("*.ipynb"):
    if ".ipynb_checkpoints" not in str(notebook_filename) and "_build" not in str(
        notebook_filename
    ):
        try:
            print(f"Executing {notebook_filename}")
            nb = nbformat.read(notebook_filename, as_version=4)
            client = NotebookClient(
                nb,
                timeout=600,
                kernel_name="python3",
                resources={"metadata": {"path": notebook_filename.parent}},
                store_widget_state=True,
            )
            client.execute()
            nbformat.write(nb, notebook_filename)
        except (CellExecutionError, CellTimeoutError):
            msg = 'Error executing the notebook "%s".\n\n' % notebook_filename
            msg += 'See notebook "%s" for the traceback.' % notebook_filename
            print(msg)
            raise
        finally:
            nbformat.write(nb, notebook_filename)



import pytest
from numpy.testing import assert_allclose

from ross.bearings.plain_journal import PlainJournal
from ross.units import Q_


@pytest.fixture
def plain_journal_perturbation():
    """Fixture for PlainJournal with perturbation method"""
    frequency = Q_([900], "RPM")
    L = Q_(10.3600055944, "in")
    oil_flow = Q_(37.86, "l/min")

    bearing = PlainJournal(
        n=3,
        axial_length=L,
        journal_radius=0.2,
        radial_clearance=1.95e-4,
        elements_circumferential=11,
        elements_axial=3,
        n_pad=2,
        pad_arc_length=176,
        preload=0,
        geometry="circular",
        reference_temperature=50,
        frequency=frequency,
        fxs_load=0,
        fys_load=-112814.91,
        groove_factor=[0.52, 0.48],
        lubricant="ISOVG32",
        sommerfeld_type=2,
        initial_guess=[0.1, -0.1],
        method="perturbation",
        operating_type="flooded",
        oil_supply_pressure=0,
        oil_flow_v=oil_flow,
    )

    return bearing


@pytest.fixture
def plain_journal_lund():
    """Fixture for PlainJournal with lund method"""
    frequency = Q_([900], "RPM")
    L = Q_(10.3600055944, "in")
    oil_flow = Q_(37.86, "l/min")

    bearing = PlainJournal(
        n=3,
        axial_length=L,
        journal_radius=0.2,
        radial_clearance=1.95e-4,
        elements_circumferential=11,
        elements_axial=3,
        n_pad=2,
        pad_arc_length=176,
        preload=0,
        geometry="circular",
        reference_temperature=50,
        frequency=frequency,
        fxs_load=0,
        fys_load=-112814.91,
        groove_factor=[0.52, 0.48],
        lubricant="ISOVG32",
        sommerfeld_type=2,
        initial_guess=[0.1, -0.1],
        method="lund",
        operating_type="flooded",
        oil_supply_pressure=0,
        oil_flow_v=oil_flow,
    )

    return bearing


def test_plain_journal_parameters_perturbation(plain_journal_perturbation):
    """Test basic parameters for perturbation method"""
    assert_allclose(plain_journal_perturbation.axial_length, 0.263144, rtol=0.0001)
    assert_allclose(plain_journal_perturbation.journal_radius, 0.2)
    assert_allclose(plain_journal_perturbation.frequency, 94.24777961)
    assert_allclose(plain_journal_perturbation.rho, 873.99629)
    assert_allclose(plain_journal_perturbation.reference_temperature, 50)


def test_plain_journal_parameters_lund(plain_journal_lund):
    """Test basic parameters for lund method"""
    assert_allclose(plain_journal_lund.axial_length, 0.263144, rtol=0.0001)
    assert_allclose(plain_journal_lund.journal_radius, 0.2)
    assert_allclose(plain_journal_lund.frequency, 94.24777961)
    assert_allclose(plain_journal_lund.rho, 873.99629)
    assert_allclose(plain_journal_lund.reference_temperature, 50)


def test_plain_journal_equilibrium_pos_perturbation(plain_journal_perturbation):
    """Test equilibrium position for perturbation method"""
    assert_allclose(
        plain_journal_perturbation.equilibrium_pos[0], 0.68733194, rtol=0.01
    )
    assert_allclose(
        plain_journal_perturbation.equilibrium_pos[1], -0.79394211, rtol=0.01
    )


def test_plain_journal_equilibrium_pos_lund(plain_journal_lund):
    """Test equilibrium position for lund method"""
    assert_allclose(plain_journal_lund.equilibrium_pos[0], 0.68733194, rtol=0.01)
    assert_allclose(plain_journal_lund.equilibrium_pos[1], -0.79394211, rtol=0.01)


def test_plain_journal_coefficients_perturbation(plain_journal_perturbation):
    """Test coefficients for perturbation method"""
    frequency = Q_(900, "RPM")
    coeffs = plain_journal_perturbation.coefficients(frequency)
    kxx, kxy, kyx, kyy = coeffs[0]
    cxx, cxy, cyx, cyy = coeffs[1]

    assert_allclose(kxx, 2497674531.1749372, rtol=0.0001)
    assert_allclose(kxy, 783937669.6587772, rtol=0.0001)
    assert_allclose(kyx, -3140562821.5290236, rtol=0.0001)
    assert_allclose(kyy, 2562440911.734241, rtol=0.0001)
    assert_allclose(cxx, 36950674.61976142, rtol=0.0001)
    assert_allclose(cxy, -37265296.2322692, rtol=0.0001)
    assert_allclose(cyx, -42642543.712838694, rtol=0.0001)
    assert_allclose(cyy, 100992315.0043159, rtol=0.0001)


def test_plain_journal_coefficients_lund(plain_journal_lund):
    """Test coefficients for lund method"""
    frequency = Q_(900, "RPM")
    coeffs = plain_journal_lund.coefficients(frequency)
    kxx, kxy, kyx, kyy = coeffs[0]
    cxx, cxy, cyx, cyy = coeffs[1]

    assert_allclose(kxx, 947508775.2790189, rtol=0.0001)
    assert_allclose(kxy, 156786732.38415018, rtol=0.0001)
    assert_allclose(kyx, -2006480985.1711535, rtol=0.0001)
    assert_allclose(kyy, 2165272905.1621084, rtol=0.0001)
    assert_allclose(cxx, 11502933.224462334, rtol=0.0001)
    assert_allclose(cxy, -13040765.779177427, rtol=0.0001)
    assert_allclose(cyx, -13051009.888004456, rtol=0.0001)
    assert_allclose(cyy, 40600798.873926796, rtol=0.0001)



import warnings
import numpy as np
from scipy.fft import fft, fftfreq

from ross.units import Q_
from ross.results import HarmonicBalanceResults


class HarmonicBalance:
    def __init__(self, rotor, n_harmonics=None):
        """
        Harmonic Balance method for nonlinear rotor dynamic systems.

        This class implements the Harmonic Balance (HB) method to compute
        steady-state responses of rotor systems under periodic excitations.
        It supports gravitational loads, external harmonic forces, and
        potential crack-induced stiffness variations.

        Parameters
        ----------
        rotor : ross.Rotor
            Rotor object representing the system model.
        n_harmonics : int, optional
            Number of harmonics to include in the analysis.
            Default is 1.
        """
        self.rotor = rotor
        self.noh = n_harmonics if n_harmonics else 1

    def run(
        self,
        forces,
        speed,
        t,
        gravity=False,
        F_ext=None,
    ):
        """
        Solve the rotor system in the frequency domain using the
        Harmonic Balance (HB) method.

        Parameters
        ----------
        forces : list of dict
            List of harmonic force definitions. Each dictionary should contain the
            following keys:

            - 'node': int
                Node index where the force is applied.

            - 'magnitudes': list, float
                List of excitation magnitudes.
                Interpretation depends on the type of excitation:
                    - For direct harmonic forces: force amplitudes (N).
                    - For unbalance-type excitations: ``m * e * speed**2`` (N),
                    where ``m`` is the unbalance mass (kg) and ``e`` is the eccentricity (m).

            - 'phases': list of float
                List of phase angles (rad).

            - 'harmonics': list of int
                List of harmonic orders (1 for fundamental, 2 for second, etc.).
        speed : float
            Rotor rotational speed (rad/s).
        t : float
            Time array used for Fourier expansion (s).
        gravity : bool, optional
            If True, include gravitational forces. Default is False.
        F_ext : ndarray, optional
            External force array of shape (ndof, N), where N is the number of time
            samples.

        Returns
        -------
        Qt : ndarray
            Complex displacement vector in frequency domain.
        Qo : ndarray
            Static (mean) displacement vector.
        dQ : ndarray
            Harmonic displacement coefficients.
        dQ_s : ndarray
            Complex conjugate of harmonic coefficients.

        Notes
        -----
        This method constructs and solves the HB system:

        .. math::
            H Q = F

        where `H` is the harmonic balance matrix and `F` is the
        assembled force vector containing both static and harmonic components.
        """
        rotor = self.rotor

        accel = 0  # Assuming always constant speed
        crack = None  # Assuming there's no crack, crack model needs to be integrated
        freq = Q_(speed, "rad/s").to("Hz").m
        dt = t[1] - t[0]

        # Harmonic force
        Fh, Fh_s = self._harmonic_force(forces)

        # Weight
        W = rotor.gravitational_force() * int(gravity)

        # External force
        Fo, Fn, Fn_s = self._external_force(dt, freq, F_ext)

        Fn += Fh
        Fn_s += Fh_s
        F = self._assemble_forces(W, Fo, Fn, Fn_s)

        # Crack stiffness matrices
        Ko, Kn, Kn_s = self._crack_stiffness_matrices(dt, freq, crack)

        # Harmonic Balance Matrix
        H = self._build_harmonic_balance_matrix(
            speed,
            accel,
            rotor.M(speed),
            rotor.K(speed),
            rotor.Ksdt(),
            rotor.C(speed),
            rotor.G(),
            Ko,
            Kn,
            Kn_s,
        )

        Qt, Qo, dQ, dQ_s = self._solve_freq_domain(H, F)

        return HarmonicBalanceResults(rotor, speed, t, Qt, Qo, dQ, dQ_s, self.noh)

    def _harmonic_force(self, forces):
        """
        Construct the harmonic force components in the frequency domain.

        Parameters
        ----------
        forces : list of dict
            List of harmonic force definitions. Each dictionary should contain the
            following keys:

            - 'node': int
                Node index where the force is applied.

            - 'magnitudes': list, float
                List of excitation magnitudes.
                Interpretation depends on the type of excitation:
                    - For direct harmonic forces: force amplitudes (N).
                    - For unbalance-type excitations: ``m * e * speed**2`` (N),
                    where ``m`` is the unbalance mass (kg) and ``e`` is the eccentricity (m).

            - 'phases': list of float
                List of phase angles (rad).

            - 'harmonics': list of int
                List of harmonic orders (1 for fundamental, 2 for second, etc.).

        Returns
        -------
        F : ndarray, complex
            Harmonic force array of shape (ndof, noh).
        F_s : ndarray, complex
            Complex conjugate of the harmonic forces.
        """
        ndof = self.rotor.ndof
        number_dof = self.rotor.number_dof

        max_harmonic = max(h for f in forces for h in f["harmonics"])
        if max_harmonic > self.noh:
            self.noh = max_harmonic

        F = np.zeros((ndof, self.noh), dtype=np.complex128)

        for f in forces:
            n = f["node"]

            for p, m, h in zip(f["phases"], f["magnitudes"], f["harmonics"]):
                cos = np.cos(p)
                sin = np.sin(p)

                Fa = m * np.array([cos, sin])
                Fb = m * np.array([-sin, cos])

                dofs = [number_dof * n, number_dof * n + 1]
                F[dofs, h - 1] += Fa - 1j * Fb

        F_s = np.conjugate(F)

        return F, F_s

    def _unbalance_force(self, node, magnitude, phase, omega, alpha=0):
        """
        Compute unbalance forces in the frequency domain.

        Parameters
        ----------
        node : list, int
            Node indices with unbalance.
        magnitude : list, float
            Unbalance magnitudes (kg·m).
        phase : list, float
            Phase angles of unbalance (rad).
        omega : float
            Angular speed (rad/s).
        alpha : float, optional
            Angular acceleration (rad/s²). Default is 0.

        Returns
        -------
        F : ndarray, complex
            Unbalance force vector.
        F_s : ndarray, complex
            Complex conjugate of the unbalance force.
        """
        ndof = self.rotor.ndof
        number_dof = self.rotor.number_dof

        F = np.zeros((ndof), dtype=np.complex128)

        for n, m, p in zip(node, magnitude, phase):
            cos = np.cos(p)
            sin = np.sin(p)

            Fa = m * omega**2 * np.array([cos, sin])
            Fa += m * alpha * np.array([-sin, cos])

            Fb = m * omega**2 * np.array([-sin, cos])
            Fb += m * alpha * np.array([cos, -sin])

            dofs = [number_dof * n, number_dof * n + 1]
            F[dofs] += Fa - 1j * Fb

        F_s = np.conjugate(F)

        return F, F_s

    def _external_force(self, dt, freq, F=None):
        """
        Compute Fourier expansion of external time-domain forces.

        Parameters
        ----------
        dt : float
            Time step (s).
        freq : float
            Fundamental excitation frequency (Hz).
        F : ndarray, optional
            External force time history of shape (ndof, N).

        Returns
        -------
        Fo : ndarray
            Static (mean) force component.
        Fn : ndarray
            Harmonic force coefficients.
        Fn_s : ndarray
            Complex conjugate of harmonic force coefficients.
        """
        ndof = self.rotor.ndof

        Fo = np.zeros(ndof, dtype=complex)
        Fn = np.zeros((ndof, self.noh), dtype=complex)
        Fn_s = np.zeros((ndof, self.noh), dtype=complex)

        if F is not None:
            dofs = list(set(np.where(F != 0)[0]))

            Fo_, Fn_ = self._Fourier_expansion(F[dofs, :], dt, freq, self.noh)

            Fo[dofs] += Fo_
            Fn[dofs, :] += Fn_
            Fn_s[dofs, :] += np.conjugate(Fn_)

        return Fo, Fn, Fn_s

    def _assemble_forces(self, W, Fo, Fn, Fn_s):
        """
        Assemble the total complex force vector for the HB system.

        Parameters
        ----------
        W : ndarray
            Static gravitational force vector.
        Fo : ndarray
            Static external force component.
        Fn : ndarray
            Harmonic force coefficients.
        Fn_s : ndarray
            Conjugate harmonic force coefficients.

        Returns
        -------
        F : ndarray, complex
            Combined force vector for harmonic balance analysis.
        """
        ndof = self.rotor.ndof

        F = np.zeros(((self.noh * 2 + 1) * ndof), dtype=complex)

        F[:ndof] = 4 * W + 2 * Fo

        for i in range(1, self.noh + 1):
            F[(2 * i - 1) * ndof : 2 * i * ndof] = 2 * Fn[:, i - 1]
            F[2 * i * ndof : (2 * i + 1) * ndof] = 2 * Fn_s[:, i - 1]

        return F

    def _crack_stiffness_matrices(self, dt, freq, crack=None):
        """
        Compute Fourier-expanded stiffness matrices for cracked shafts.

        Parameters
        ----------
        dt : float
            Time step (s).
        freq : float
            Rotational frequency (Hz).
        crack : object, optional
            Crack model object providing `dofs`, `crack_coeff`, and `_Kflex()`
            methods.

        Returns
        -------
        Ko : ndarray
            Static stiffness correction matrix.
        Kn : ndarray
            Harmonic stiffness correction matrices.
        Kn_s : ndarray
            Conjugate harmonic stiffness matrices.
        """
        ndof = self.rotor.ndof
        n_aux = 2 * self.noh

        Ko = np.zeros((ndof, ndof), dtype=complex)
        Kn = np.zeros((ndof, ndof, n_aux), dtype=complex)
        Kn_s = np.zeros((ndof, ndof, n_aux), dtype=complex)

        if crack:
            dof = crack.dofs
            crack_coeff = crack.crack_coeff

            Kco, Kcn, Kcn_s = self._Fourier_expansion(crack_coeff, dt, freq, n_aux)

            Ko[dof, dof] = crack._Kflex(Kco.reshape(-1, 1))[:, :, 0]
            Kn_ = crack._Kflex(Kcn)
            Kn_s_ = crack._Kflex(Kcn_s)

            for i in range(n_aux):
                Kn[dof, dof, i] = Kn_[:, :, i]
                Kn_s[dof, dof, i] = Kn_s_[:, :, i]

        return Ko, Kn, Kn_s

    def _build_harmonic_balance_matrix(
        self,
        speed,
        accel,
        M,
        K,
        Ksdt,
        C,
        G,
        Ko,
        Kn,
        Kn_s,
        Co=None,
        Cn=None,
        Cn_s=None,
    ):
        """
        Construct the Harmonic Balance matrix `H`.

        Parameters
        ----------
        speed : float
            Rotational speed (rad/s).
        accel : float
            Angular acceleration (rad/s²).
        M, K, Ksdt, C, G : ndarray
            Rotor mass, stiffness, stiffness time-derivative, damping, and
            gyroscopic matrices.
        Ko : ndarray
            Static crack stiffness correction.
        Kn, Kn_s : ndarray
            Harmonic stiffness correction and its conjugate.
        Co, Cn, Cn_s : ndarray, optional
            Damping correction matrices. Default: zeros.

        Returns
        -------
        H0 : ndarray, complex
            Full harmonic balance system matrix.
        """
        ndof = self.rotor.ndof
        noh = self.noh

        # Co, Cn, Cn_s are already considered in C matrix (bearing elements)
        Co = np.zeros((ndof, ndof), dtype=complex)
        Cn = np.zeros((ndof, ndof, 2 * noh), dtype=complex)
        Cn_s = np.zeros((ndof, ndof, 2 * noh), dtype=complex)

        # alpha and beta are already considered in C matrix (shaft elements)
        alpha = 0
        beta = 0

        size = ndof * (2 * noh + 1)
        H0 = np.zeros((size, size), dtype=complex)

        idx0 = slice(0, ndof)
        idx1 = slice(1 * ndof, 2 * ndof)
        idx2 = slice(2 * ndof, 3 * ndof)

        H0[idx0, idx0] = Ko + 2 * (K + Ksdt * accel)

        K_aux = 1j * beta * speed
        aux_C = 1j * speed
        aux1 = 1 * K_aux + 1

        H0[idx1, idx2] = np.conjugate(aux1) * Kn[:, :, 1] - 1 * aux_C * Cn[:, :, 1]
        H0[idx2, idx1] = aux1 * Kn_s[:, :, 1] + 1 * aux_C * Cn_s[:, :, 1]

        for n in range(1, noh + 1):
            idx3 = slice((2 * n - 1) * ndof, 2 * n * ndof)
            idx4 = slice(2 * n * ndof, (2 * n + 1) * ndof)

            aux2 = -2 * n**2 * speed**2 + 2 * n * aux_C * alpha
            aux3 = 2 * n * 1j * speed**2
            aux4 = n * aux_C
            aux5 = 2 * n * K_aux + 2
            aux6 = 1 * n * K_aux + 1
            H0[idx3, idx3] = (
                aux2 * M
                + aux3 * G
                + aux4 * (2 * C + Co)
                + aux5 * K
                + 2 * (Ksdt * accel)
                + aux6 * Ko
            )

            H0[idx4, idx4] = (
                np.conjugate(aux2) * M
                + np.conjugate(aux3) * G
                + np.conjugate(aux4) * (2 * C + Co)
                + np.conjugate(aux5) * K
                + 2 * (Ksdt * accel)
                + np.conjugate(aux6) * Ko
            )

            H0[idx3, idx0] = Kn[:, :, n - 1]
            H0[idx4, idx0] = Kn_s[:, :, n - 1]
            H0[idx0, idx3] = aux6 * Kn_s[:, :, n - 1] + aux4 * Cn_s[:, :, n - 1]
            H0[idx0, idx4] = (
                np.conjugate(aux6) * Kn[:, :, n - 1]
                + np.conjugate(aux4) * Cn[:, :, n - 1]
            )

            if n < noh:
                aux7 = 2 * noh - 2 * n + 2

                for k in range(2, aux7):
                    idx5 = slice((k - 1) * ndof, k * ndof)
                    idx6 = slice((k - 1 + 2 * n) * ndof, (k + 2 * n) * ndof)
                    aux8 = np.ceil((k + 2 * n - 1) / 2)

                    if np.mod(k, 2) == 1:
                        A_ = (-aux8 * K_aux + 1) * Kn[:, :, n - 1] - aux8 * aux_C * Cn[
                            :, :, n - 1
                        ]
                    else:
                        A_ = (aux8 * K_aux + 1) * Kn_s[
                            :, :, n - 1
                        ] + aux8 * aux_C * Cn_s[:, :, n - 1]

                    H0[idx5, idx6] = A_

                aux9 = 2 * n + 1
                for k in range(2 * noh + 1, aux9, -1):
                    idx7 = slice((k - 1) * ndof, k * ndof)
                    idx8 = slice((k - 1 - 2 * n) * ndof, (k - 2 * n) * ndof)
                    aux10 = np.ceil((k - 2 * n - 1) / 2)

                    if np.mod(k, 2) == 1:
                        B_ = (-aux10 * K_aux + 1) * Kn_s[
                            :, :, n - 1
                        ] - aux10 * aux_C * Cn_s[:, :, n - 1]
                    else:
                        B_ = (+aux10 * K_aux + 1) * Kn[
                            :, :, n - 1
                        ] + aux10 * aux_C * Cn[:, :, n - 1]

                    H0[idx7, idx8] = B_

        aux11 = 2 * noh - 1
        for n in range(2 * noh, 2, -1):
            aux12 = 2 * n - 1

            if aux12 > 2 * noh:
                aux12 = 2 * noh + 1

            if aux11 < 2:
                aux11 = 1

            aux13 = 2 * n + 1

            for k in range(aux12, aux11, -1):
                idx9 = slice((k - 1) * ndof, k * ndof)
                idx10 = slice((aux13 - k - 1) * ndof, (aux13 - k) * ndof)
                aux14 = np.ceil((aux13 - k - 1) / 2)

                if np.mod(k, 2) == 1:
                    C_ = (+aux14 * K_aux + 1) * Kn_s[
                        :, :, n - 1
                    ] + aux14 * aux_C * Cn_s[:, :, n - 1]
                else:
                    C_ = (-aux14 * K_aux + 1) * Kn[:, :, n - 1] - aux14 * aux_C * Cn[
                        :, :, n - 1
                    ]

                H0[idx9, idx10] = C_

            aux11 = aux11 - 2

        return H0

    def _solve_freq_domain(self, H, F):
        """
        Solve the Harmonic Balance system in frequency domain.

        Parameters
        ----------
        H : ndarray
            Harmonic Balance matrix.
        F : ndarray
            Force vector.

        Returns
        -------
        Qt : ndarray, complex
            Full frequency-domain displacement vector.
        Qo : ndarray, complex
            Static displacement component.
        Qn : ndarray, complex
            Harmonic displacement components.
        Qn_s : ndarray, complex
            Conjugate harmonic displacement components.

        Notes
        -----
        If the system is singular, the pseudo-inverse is used instead of direct solve.
        """
        ndof = self.rotor.ndof

        try:
            Qt = np.linalg.solve(H, F)
        except np.linalg.LinAlgError as err:
            warnings.warn(
                f"{err} error. Using the pseudo-inverse to proceed.", UserWarning
            )
            Qt = F @ np.linalg.pinv(H)

        Qo = np.real(Qt[:ndof])
        Qn = np.zeros((ndof, self.noh), dtype=complex)
        Qn_s = np.zeros((ndof, self.noh), dtype=complex)

        for i in range(1, self.noh + 1):
            Qn[:ndof, i - 1] = Qt[(2 * i - 1) * ndof : (2 * i) * ndof]
            Qn_s[:ndof, i - 1] = Qt[(2 * i) * ndof : (2 * i + 1) * ndof]

        return Qt, Qo, Qn, Qn_s

    @staticmethod
    def _reconstruct_time_domain(omega, t, Qo, dQ, n_harmonics):
        """
        Reconstruct the time-domain response from frequency-domain results.

        Parameters
        ----------
        omega : float
            Rotational speed (rad/s).
        t : array_like
            Time vector (s).
        Qo : ndarray
            Static displacement vector.
        dQ : ndarray
            Harmonic displacement coefficients.
        n_harmonics : int, optional
            Number of harmonics to include in the analysis.

        Returns
        -------
        y : ndarray
            Displacement response over time.
        ydot : ndarray
            Velocity response over time.
        y2dot : ndarray
            Acceleration response over time.
        """
        shape = (np.size(Qo), len(t))

        sum_y = np.zeros(shape)
        sum_ydot = np.zeros(shape)
        sum_y2dot = np.zeros(shape)

        for i in range(1, n_harmonics + 1):
            an = np.transpose(np.array([np.real(dQ[:, i - 1])]))
            bn = np.transpose(np.array([-np.imag(dQ[:, i - 1])]))

            cos = np.array([np.cos(i * omega * t)])
            sin = np.array([np.sin(i * omega * t)])

            sum_y += np.dot(an, cos) + np.dot(bn, sin)
            sum_ydot += i * omega * (np.dot(bn, cos) - np.dot(an, sin))
            sum_y2dot -= (i * omega) ** 2 * (np.dot(an, cos) + np.dot(bn, sin))

        y = Qo[:, np.newaxis] / 2 + sum_y
        ydot = sum_ydot
        y2dot = sum_y2dot

        return y, ydot, y2dot

    @staticmethod
    def _Fourier_expansion(F, dt, fo, size):
        """
        Perform Fourier expansion of a time-domain signal.

        Parameters
        ----------
        F : ndarray
            Input force array of shape (ndof, N).
        dt : float
            Time step (s).
        fo : float
            Fundamental frequency (Hz).
        size : int
            Number of harmonic components to compute.

        Returns
        -------
        Fo : ndarray
            Mean (static) component.
        Fn : ndarray
            Complex harmonic coefficients.

        Notes
        -----
        Uses FFT to compute coefficients corresponding to multiples of fundamental
        frequency.
        """
        row, N = F.shape
        b = N // 2

        X = fft(F)[:, :b]
        X *= 2 / N
        freqs = fftfreq(N, dt)[:b]

        Fo = np.real(X[:, 0])
        an = np.real(X)
        bn = -np.imag(X)

        Fn = np.zeros((row, size), dtype=complex)
        for n in range(1, size + 1):
            idx = np.argmin(np.abs(freqs - n * fo))
            Fn[:, n - 1] = an[:, idx] - 1j * bn[:, idx]

        return Fo, Fn
